import aio
import inspect
import sys
import threading as __threading__

sys.modules["__threading__"] = __threading__
from __threading__ import *

# mark not started but no error
aio.error = None

aio.paused = False
aio.fd = {}
aio.pstab = {}


def _shutdown():
    print(__file__, "_shutdown")


# https://docs.python.org/3/library/threading.html#threading.excepthook

# a green thread
# FIXME: fix wapy BUG 882 so target can be None too in preempt mode

# TODO: default granularity with https://docs.python.org/3/library/sys.html#sys.setswitchinterval


def excepthook(*argv, **kw):
    print("24 threading.excepthook", __file__, argv, kw)


class _dangling:
    @classmethod
    def copy(cls):
        # __threading__._MainThread()
        return set([])

    @classmethod
    def clear(cls):
        pass

    @classmethod
    def update(cls, saved):
        pass


class Local:
    pass


def local():
    return Local


class Lock:
    count = 0

    def __enter__(self):
        self.acquire()

    def __exit__(self, *tb):
        self.release()

    def acquire(self, blocking=True, timeout=-1):
        self.count += 1
        return True

    def release(self):
        self.count -= 1

    def locked(self):
        return self.count > 0


RLock = Lock


class Condition:
    def __init__(self, lock=None):
        self.lock = lock or Lock()

    def acquire(self, *args):
        return self.lock.acquire()

    def release(self):
        self.lock.release()

    def wait(self, timeout=None):
        raise RuntimeError("notify not supported")

    def wait_for(self, predicate, timeout=None):
        raise RuntimeError("wait not supported")


class Thread:
    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}, *, daemon=None):
        # def __init__(self, group=None, target=None, name=None, args=(), kwargs={}):
        self.args = args
        self.kwargs = kwargs
        self.name = name
        self.slice = 0
        self.last = aio.rtclock()

        if target:
            if hasattr(target, "run"):
                if name is None:
                    self.name = name or target.__class__.__name__
                self.run = target.run
            else:
                self.run = target

            if name is None:
                try:
                    self.name = "%s-%s" % (self.run.__name__, id(self))
                except:
                    pass
        else:
            target = self

        if self.name is None:
            self.name = "%s-%s" % (self.__class__.__name__, id(self))
        self.status = None

    async def wrap(self):
        for idle in self.run(*self.args, **self.kwargs):
            await aio.sleep(0)

    async def runner(self, coro):
        self.status = True
        try:
            # TODO: pass thread local context here
            async with aio.ctx(self.slice).call(coro):
                self.status = False
        except Exception as e:
            self.status = repr(e)
            sys.print_exception(e, sys.stderr)

    if __UPY__:

        def __iter__(self):
            if self.status is True:
                rtc = aio.rtclock()
                self.delta = (rtc - self.last) - self.slice
                if self.delta < 0:
                    self.delta = 0
                yield from aio.sleep_ms(self.slice - int(self.delta / 2))
                self.last = rtc

        __await__ = __iter__
    else:

        def __await__(self):
            if self.status is True:
                rtc = aio.rtclock()
                self.delta = (rtc - self.last) - self.slice
                if self.delta < 0:
                    self.delta = 0
                # no sleep_ms on cpy
                yield from aio.sleep_ms(float(self.slice - int(self.delta / 2)) / 1_000).__await__()
                # return aio.sleep( float(self.slice - int(self.delta / 2)) / 1_000 )
                self.last = rtc

    def rt(self, slice):
        self.slice = int(float(slice) * 1_000)
        return self

    def start(self):
        aio.pstab.setdefault(self.name, [])
        if self.run:
            if not inspect.iscoroutinefunction(self.run):
                self.status = True
                aio.create_task(self.wrap())
            else:
                coro = self.run(*self.args, **self.kwargs)
                pdb("168:", self.name, "starting", coro)
                aio.create_task(self.runner(coro))
            aio.pstab[self.name].append(self)

        return self

    def join(self):
        embed.enable_irq()
        while self.is_alive():
            aio_suspend()
        embed.disable_irq()

    def __bool__(self):
        return self.is_alive() and not aio.exit

    def is_alive(self):
        return self.status is True


class Timer:
    def __init__(self, interval, function, args=None, kwargs=None):
        self.abort = False
        self.interval = interval
        self.thread = Thread(group=None, target=self.function, args=args, kwargs=kwargs)

    async def defer(self):
        await asyncio.sleep(self.interval)
        if not self.abort:
            self.thread.start()
        del self.abort, self.interval, self.thread

    def start(self):
        aio.create_task(self.defer())
        return self

    def cancel(self):
        self.abort = True


def service(srv, *argv, **kw):
    embed.log(f"starting green thread : {srv}")
    thr = aio.Thread(group=None, target=srv, args=argv, kwargs=kw).start()
    srv.__await__ = thr.__await__
    return aio.pstab.setdefault(srv, thr)


aio.task = service


def proc(srv):
    return aio.pstab.get(srv)


def _register_atexit(proc):
    print(__name__, "_register_atexit not implemented")


class Runnable:
    def __await__(self):
        yield from aio.pstab.get(self).__await__()


# replace with green threading
import sys

sys.modules["threading"] = sys.modules["aio.gthread"]
sys.modules["dummy_threading"] = sys.modules["aio.gthread"]



import sys
import io

if not __UPY__:
    import socket

    socket.setdefaulttimeout(0.0)
else:
    print("7: usocket implementation required")


import os  # unlink


import aio
import platform

temporary = []


async def aio_sock_open(sock, host, port):
    while True:
        try:
            sock.connect(
                (
                    host,
                    port,
                )
            )
        except BlockingIOError:
            await aio.sleep(0)
        except OSError as e:
            # 30 emsdk, 106 linux
            if e.errno in (30, 106):
                return sock
            sys.print_exception(e)


def fix_url(maybe_url):
    url = str(maybe_url)
    if url.startswith("http://"):
        pass
    elif url.startswith("https://"):
        pass
    elif url.startswith("https:/"):
        url = "https:/" + url[6:]
    elif url.startswith("http:/"):
        url = "http:/" + url[5:]
    return url


def mktemp(suffix=""):
    global temporary
    tmpname = f"/tmp/tmp-{aio.ticks}-{len(temporary)}{suffix}"
    temporary.append(tmpname)
    return tmpname


class fopen:
    if __WASM__ and hasattr(platform, "ffi"):
        flags = platform.ffi(
            {
                "redirect": "follow",
                "credentials": "omit",
            }
        )
    else:
        print("69: platform has no object serializer")
        flags = {}

    def __init__(self, maybe_url, mode="r", flags=None, encoding="UTF-8", errors=None, newline=None, closefd=True, opener=None):
        self.url = fix_url(maybe_url)
        self.mode = mode
        flags = flags or self.__class__.flags
        print(f'76: fopen: fetching "{self.url}"')

        self.tmpfile = None

    async def __aexit__(self, *exc):
        if self.tmpfile:
            self.filelike.close()
            try:
                os.unlink(self.tmpfile)
            except FileNotFoundError as e:
                print("78: Async I/O error : file not found", self.url)
        del self.filelike, self.url, self.mode, self.tmpfile

    if __WASM__:

        async def __aenter__(self):
            import platform

            self.tmpfile = shell.mktemp()
            cf = platform.window.cross_file(self.url, self.tmpfile, self.flags)
            try:
                content = await platform.jsiter(cf)
            except Exception as e:
                print("91:", e)

            if "b" in self.mode:
                self.filelike = open(content, "rb")
            else:
                self.filelike = open(content, "r", encoding="utf-8")

            self.filelike.path = content

            def rename_to(target):
                # will be closed
                self.filelike.close()
                os.rename(self.tmpfile, target)
                self.tmpfile = None
                del self.filelike
                return target

            self.filelike.rename_to = rename_to
            return self.filelike

    else:

        async def __aenter__(self):
            import aiohttp

            async with aiohttp.ClientSession() as session:
                async with session.get(self.url) as response:
                    if response.status != 200:
                        raise FileNotFoundError(self.url)

                    # For large files use response.content.read(chunk_size) instead.
                    if "b" in self.mode:
                        self.filelike = io.BytesIO(await response.read())
                    else:
                        self.filelike = io.StringIO((await response.read()).decode())

            return self.filelike


class sopen:
    def __init__(self, url, mode, tmout):  # =6):
        self.host, port = url.rsplit(":", 1)
        self.port = int(port)
        if __WASM__ and __import__("platform").is_browser:
            if not url.startswith("ws"):
                pdb(f"switching to {self.port}+20000 as websocket")
                self.port += 20000

        self.sock = socket.socket()
        # self.sock.setblocking(0)

    # overload socket ?

    def fileno(self):
        return self.sock.fileno()

    def send(self, *argv, **kw):
        self.sock.send(*argv, **kw)

    def recv(self, *argv):
        return self.sock.recv(*argv)

    # ============== specific ===========================

    async def __aenter__(self):
        # use async
        await aio_sock_open(self.sock, self.host, self.port)
        return self

    async def __aexit__(self, exc_type, exc, tb):
        aio.protect.append(self)
        aio.defer(self.sock.close, (), {})
        del self.port, self.host, self.sock

    def read(self, size=-1):
        return self.recv(0)

    def write(self, data):
        if isinstance(data, str):
            return self.sock.send(data.encode())
        return self.sock.send(data)

    def print(self, *argv, **kw):
        # use EOL
        # kw.setdefault("end","\r\n")
        kw["file"] = io.StringIO(newline="\r\n")
        print(*argv, **kw)
        self.write(kw["file"].getvalue())

    def __enter__(url, mode, tmout):
        # use softrt (wapy)
        return aio.await_for(self.__aenter__())

    def __exit__(self, exc_type, exc, tb):
        # self.sock.close()
        pass



import sys
import asyncio

asyncrun = asyncio.run

from .__init__ import VERSION

print(f" *pygbag {VERSION}*")

from pathlib import Path


async def import_site(sourcefile=None, simulator=False, async_input=None, async_pkg=None):
    import sys
    import os
    from pathlib import Path

    if ("--sim" not in sys.argv) and ("--piny" not in sys.argv) and not simulator:
        from .app import main_run, set_args

        app_folder, mainscript = set_args(sys.argv[-1])
        # run pygbag build/server
        await main_run(app_folder, mainscript)
        return True

    # or run as a native simulator

    mod_dir = Path(__file__).parent
    support = mod_dir / "support"
    cur_dir = os.getcwd()
    print(
        f"""
            - single thread no-os wasm simulator -

    module directory : {mod_dir}
    platform support : {support}
    {sys.argv=}
    {sourcefile=}
    {cur_dir=}

"""
    )

    import sys, os, builtins

    # sys.path.append(str(support / "cross"))

    # need them earlier than aio

    def pdb(*argv):
        print(*argv, file=sys.__stderr__)

    builtins.pdb = pdb

    def print_exception(e, out=sys.stderr, **kw):
        kw["file"] = out
        traceback.print_exc(**kw)

    # abstract placeholder for missing host objects
    class NoOp(object):
        def __init__(self, *argv, **env):
            self.__descr = " ".join(map(str, argv))
            self.__lastc = "__init__"

        def __call__(self, *argv, **env):
            print(self, "call:", argv, env)
            return None

        def __nop(self, *argv, **env):
            print(f"{self}.{self.__lastc}(*{argv}, **{env})")
            return self

        def __nonzero__(self):
            return 0

        def __nop__(self, other):
            return 0

        __add__ = __iadd__ = __sub__ = __isub__ = __mul__ = __imul__ = __div__ = __idiv__ = __nop__

        def __getattr__(self, attr):
            self.__lastc = attr
            return self.__nop

        def iteritems(self):
            print(self, "iterator")
            return []

        def __del__(self):
            pass

        def __repr__(self):
            return "\nNoOp: %s" % self.__descr

        __str__ = __repr__

    class window_navigator:
        userAgent = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36"

    # fake host document.window
    import platform as fakehost

    def truc(*argv, **kw):
        print("truc", argv, kw)

    def prompt():
        # FIXME for js style console should be
        # CSI(f"{TTY.LINES+TTY.CONSOLE};1H{prompt}")
        print("\r>X> ", end="")

    fakehost.is_browser = False
    fakehost.async_input = async_input
    fakehost.window = NoOp("platform.window")
    fakehost.window.console = NoOp("platform.window.console")
    fakehost.window.console.log = print
    fakehost.window.get_terminal_console = truc
    fakehost.window.RAW_MODE = 0
    fakehost.window.navigator = window_navigator
    fakehost.readline = lambda: ""
    fakehost.prompt = prompt

    def set_raw_mode(mode):
        import platform as fakehost

        fakehost.window.RAW_MODE = mode

    fakehost.window.set_raw_mode = set_raw_mode
    fakehost.window.set_raw_mode(0)

    import aio.filelike

    fakehost.fopen = aio.filelike.fopen

    # cannot fake a wasm cpu but fake the platform AND the module

    sys.platform = "emscripten"

    class fake_EventTarget:
        clients = {}
        events = []

        async def process(self):
            ...

    def patch_platform_system():
        return "Emscripten"

    # et = EventTarget()
    class __EMSCRIPTEN__(object):
        EventTarget = fake_EventTarget()
        ticks = 0

        def __init__(self):
            import platform

            platform.system = patch_platform_system
            self.platform = platform

        def __getattribute__(self, name):
            try:
                return object.__getattribute__(self, name)
            except:
                return object.__getattribute__(self.platform, name)

        @classmethod
        def PyConfig_InitPythonConfig(*argv):
            pass

        @classmethod
        def init_platform(*argv):
            pass

        @classmethod
        def flush(cls):
            sys.stdout.flush()
            sys.stderr.flush()

        @classmethod
        def system(cls):
            return "Linux"

        @classmethod
        def no_op(cls, *argv, **kw):
            ...

        run = no_op

        set_ps1 = no_op

        set_ps2 = no_op

        prompt = no_op

        is_browser = False

        js = pdb
        run_script = pdb

    __EMSCRIPTEN__ = __EMSCRIPTEN__()

    sys.modules["__EMSCRIPTEN__"] = __EMSCRIPTEN__
    sys.modules["embed"] = __EMSCRIPTEN__

    print(" =============== pythonrc =================")
    with open(support / "pythonrc.py", "r") as file:
        exec(file.read(), globals(), globals())

    print(" =============== /pythonrc =================")

    import zipfile
    import aio.toplevel
    import ast
    from pathlib import Path

    class TopLevel_async_handler(aio.toplevel.AsyncInteractiveConsole):
        HTML_MARK = '"' * 3 + " # BEGIN -->"

        #        @classmethod
        #        async def async_imports(cls, callback, *wanted, **kw):
        #            ...

        @classmethod
        def list_imports(cls, code=None, file=None, hint=""):
            return []

        def eval(self, source):
            for count, line in enumerate(source.split("\n")):
                if not count:
                    if line.startswith("<"):
                        self.buffer.append(f"#{line}")
                        continue
                self.buffer.append(line)

            if count:
                self.line = None
                self.buffer.insert(0, "#")
            print(f"178: {count} lines queued for async eval")

        async def async_get_pkg(cls, want, ex, resume):
            print(
                """
224:TODO: PEP723
    async def async_get_pkg(cls, want, ex, resume)
"""
            )

    # start async top level machinery and add a console.
    await TopLevel_async_handler.start_toplevel(platform.shell, console=True)

    # ns = vars(__import__(__name__))

    # ns["TopLevel_async_handler"] = TopLevel_async_handler

    aio.toplevel.handler = TopLevel_async_handler

    sourcefile = sourcefile or str(sys.argv[-1])

    __import__("__main__").__file__ = sourcefile

    await async_pkg

    if "--piny" in sys.argv:
        from . import mutator

        mutator.transform_file(sourcefile, f"{sourcefile[:-3]}.pn")

    else:
        import aio.clock

        # asyncio.create_task( aio.clock.loop() )
        aio.clock.start(x=80)

        print(__name__, "sim repl ready for", sourcefile)

        if sys.path[0] != cur_dir:
            sys.path.insert(0, cur_dir)

        await shell.runpy(sourcefile)
        shell.interactive()

        # if you don't reach that step
        # your main.py has an infinite sync loop somewhere !
        print(f"{platform.is_browser=}, sim is running")

        while not aio.exit:
            await aio.sleep(0.016)
        print(__name__, "sim terminating")


if __name__ == "__main__":
    asyncio.run(import_site())


#



import asyncio
import sys

# https://www.panda3d.org/reference/cxx/coordinateSystem_8h_source.html
# renderman CS_yup_left,  // Y-Up, Left-handed
# https://www.evl.uic.edu/ralph/508S98/coordinates.html


out = sys.__stdout__.write


import os

# for Rich color output
os.environ["FORCE_COLOR"] = "1"
damages = {}
evpos = {}
clog = []


# def ESC(*argv, flush=True):
#    for arg in argv:
#        sys.__stdout__.write(chr(0x1B))
#        sys.__stdout__.write(arg)
#    if flush:
#        sys.stdout.flush()


# def CSI(*argv):
#    for arg in argv:
#        ESC(f"[{arg}", flush=False)
#    sys.stdout.flush()


# slots = 'x y z t width depth height event'
# slots = slots.split(' ')
try:
    const
except:
    const = lambda x: x

X = const(0)
# Y = const(1)
Z = const(2)
# T = const(3)
# W = const(4)
# D = const(5)
# H = const(6)
# E = const(7)


class NodePath:
    root = None
    count = 0

    def __init__(self, parent, node, pos=None, **kw):
        if parent:
            assert isinstance(parent, NodePath)
        # FIXME: self.parent = weakref.ref(parent) <= no weakref yet in micropython
        self.parent = parent
        self.node = node
        self.pos = pos or [0, 0, 0]
        self.name = kw.pop("name", "n_{0:03d}".format(self.__class__.count, 3))
        self.dirty = False
        self.children = []
        self.__class__.count += 1

        if parent:  # and self.root!=parent:
            if not self in parent.children:
                parent.children.append(self)


def npos(n):
    if isinstance(n, NodePath):
        return n.pos
    return n


def get_x(n):
    return npos(n)[X]


def get_z(n):
    return npos(n)[Z]


def set_any(np, v, slot):
    global rd
    v = int(v)
    ov = npos(np)[slot]
    npos(np)[slot] = v
    if isinstance(np, NodePath) and v != ov:
        render.dirty = np.dirty = True


def set_x(n, v):
    set_any(n, v, X)


def set_z(n, v):
    set_any(n, v, Z)


def set_text(np, t):
    t = str(t)
    ot = np.node.text
    if ot != t:
        np.node.text = t
        render.dirty = np.dirty = True


import select
import os
from _xterm_parser.events import MouseEvent, Key

buffer_console = []

# import shutil
import os


class TTY:
    stdin = 0
    closed = False

    COLUMNS, _ = os.get_terminal_size()
    LINES = 25

    try:
        CONSOLE = os.get_console_size()
    except:
        CONSOLE = 32 - LINES

    echo = True
    console = True

    prompts = []

    readline = []

    line = ""

    # cursor pos
    L = 1
    C = 1

    event_type = ""
    last_event_type = "load"

    event_data = ""
    last_event_data = "window"

    last_state = False

    @classmethod
    def prompt(self):
        if len(self.prompts):
            self.prompts.clear()
            import platform

            print("\r>>> ", end="")
            self.flush()

    @classmethod
    def set_raw(self, state):
        import platform
        import termios, tty

        if state:
            stdin = sys.stdin.fileno()
            self.stdin = stdin

            ESC("7")

            self.attrs_before = termios.tcgetattr(stdin)
            attrs_raw = termios.tcgetattr(stdin)
            attrs_raw[tty.LFLAG] &= ~(termios.ECHO | termios.ICANON | termios.IEXTEN | termios.ISIG)
            attrs_raw[tty.IFLAG] &= ~(termios.IXON | termios.IXOFF | termios.ICRNL | termios.INLCR | termios.IGNCR)
            attrs_raw[tty.CC][termios.VMIN] = 1

            # keyb
            termios.tcsetattr(stdin, termios.TCSANOW, attrs_raw)

            # mouse
            CSI("?1000h", "?1003h", "?1015h", "?1006h")

            try:
                # platform.window.set_raw_mode(1)
                ...
            except:
                ...
        else:
            try:
                platform.window.set_raw_mode(0)
            except:
                ...

            CSI("?1000l", "?1003l", "?1015l", "?1006l")

            try:
                termios.tcsetattr(self.stdin, termios.TCSANOW, self.attrs_before)
            except:
                ...

            # get back full scrolling
            CSI("999;999r")
            # restore cur+attr
            ESC("8")

    if sys.platform in ("emscripten", "wasi") and not aio.cross.simulator:
        flush = __import__("embed").flush
    else:
        flush = sys.__stdout__.flush

    @classmethod
    def handler_console_input(self, ev):
        global clog
        if ev.is_printable:
            if self.echo:
                print(ev.character, end="")
                sys.__stdout__.flush()
            buffer_console.append(ev.character)

        elif ev.name == "enter":
            # stdin console injection. bypass embed.readline() see pythonrc.py
            print(end="\r\n")
            aio.toplevel.handler.instance.buffer.append("".join(buffer_console))
            TTY.prompts.append(1)
            buffer_console.clear()

        elif ev.name == "ctrl_d":
            # aio.exit_now(0)
            self.closed = True
        elif ev.name == "ctrl_l":
            clear(LINES=-1, prompt=">-> ")
        else:
            # clog.append(repr(list(ev.__rich_repr__())))
            ...

    @classmethod
    def handler_readline_input(self, ev):
        global clog

        # virtual gamepad handler

        if not TTY.console:
            if ev.name in ("space"):
                import pygame

                uev = pygame.event.Event(pygame.USEREVENT + 2)
                pygame.event.post(uev)
                return False
            if ev.name in ("left", "right", "up", "down"):
                if ev.name == "left":
                    return vpad.emit(vpad.X, -1.0)

                if ev.name == "right":
                    return vpad.emit(vpad.X, +1.0)

                if ev.name == "up":
                    return vpad.emit(vpad.Z, +1.0)

                if ev.name == "down":
                    return vpad.emit(vpad.Z, -1.0)

        # await input() handler
        if ev.is_printable:
            self.readline.append(ev.character)
            return False

        # validate input()

        if ev.name == "enter":
            self.last_list = self.line
            self.line = "".join(self.readline)
            self.readline.clear()
            return True
        else:
            # clog.append(repr(list(ev.__rich_repr__())))
            ...
        return False

    @classmethod
    def handler_events(self, ev):
        global evpos

        for ez in range(ev.y - 1, ev.y + 2):
            if evpos.get(ez, None):
                for k in evpos[ez]:
                    if not len(evpos[ez][k]):
                        continue
                    # clog.append( repr( evpos[ez][k] ) )
                    for slot in evpos[ez][k]:
                        xl, xr, tag = slot
                        if (xl < ev.x) and (ev.x < xr):
                            if ev.button:
                                self.event_type = "click"
                            else:
                                self.event_type = "hover"
                            self.event_data = tag
                            return

    @classmethod
    def input(self):
        global clog, parser, buffer_console
        rl_complete = False
        if self.event_type:
            self.last_event_type = self.event_type

        if self.event_data:
            self.last_event_data = self.event_data

        self.event_type = ""
        self.event_data = ""
        self.state_changed = False

        if select.select([self.stdin], [], [], 0)[0]:
            try:
                payload = os.read(self.stdin, 1024)
                if payload:
                    for event in parser.feed(payload.decode("utf-8")):
                        if isinstance(event, MouseEvent):
                            self.L = event.x
                            self.C = event.y
                            TTY.console = event.y > TTY.LINES
                            if self.last_state != TTY.console:
                                self.last_state = TTY.console
                                self.state_changed = True
                            # only last event will stay.
                            self.handler_events(event)

                        elif isinstance(event, Key):
                            if TTY.console:
                                self.handler_console_input(event)
                            else:
                                rl_complete = self.handler_readline_input(event)
                                # clog.append( repr(list(event.__rich_repr__())) )
            except OSError:
                ...

        if rl_complete:
            return self.line
        return ""


def goto_xz(x, z):
    CSI("%d;%dH" % (z, x))


# cleareos ED0          Clear screen from cursor down          ^[[J
# cleareos ED0          Clear screen from cursor down          ^[[0J
# clearbos ED1          Clear screen from cursor up            ^[[1J
# clearscreen ED2       Clear entire screen                    ^[[2J


def clear(LINES=0, CONSOLE=0, prompt=""):
    global damages

    TTY.LINES = max(TTY.LINES, LINES)

    TTY.CONSOLE = max(TTY.CONSOLE, CONSOLE)

    if CONSOLE > 0:
        # damage whole zone
        for x in damages:
            damages[x] = 0
            if x > TTY.LINES:
                break

        goto_xz(1, TTY.LINES + 1)
        CSI("0J", "1J", f"{TTY.LINES+1};{TTY.LINES+TTY.CONSOLE}r", f"{TTY.LINES+TTY.CONSOLE-1};1H{prompt}")

    elif CONSOLE < 0:
        goto_xz(1, TTY.LINES + 1)
        CSI(f"0J", f"{TTY.LINES+1};1H{prompt}")
    else:
        # damage whole screen
        for x in damages:
            damages[x] = 0
        # clear all events
        evpos.clear()
        CSI("2J")


class Node:
    def __init__(self, text=None, **kw):
        self.text = text
        self.set_filter(lambda x: x)
        self.clip = False

    def set_filter(self, flt):
        self.flt = flt

    def filter(self, np):
        return self.flt(self.text)


class render(NodePath):
    DX = const(1)
    DZ = const(40)

    IX = const(1)
    IZ = const(-1)

    # wr0 = '\x1b%d\x1b[?25%s'
    # wr = sys.stdout.write
    @classmethod
    def wr(cls, data):
        print(data, end="")

    def __init__(self):
        self.__class__.root = self
        self.dirty = False
        NodePath.__init__(self, None, self)

    def __enter__(self):
        # self.wr(self.wr0 % ( 7 , 'l' ) )
        self.wr("\x1b7\x1b[?25l")
        return self

    def __exit__(self, *tb):
        # self.wr( self.wr0 % ( 8, 'h' ) )
        self.wr("\x1b8\x1b[?25h")

    @classmethod
    def draw_child(cls, parent, np, lvl):
        pos = npos(np)
        x = cls.DX + (cls.IX * pos[0])
        z = cls.DZ + (cls.IZ * pos[2])
        cls.wr("\x1b[%d;%dH[ %s : %s ]  " % (z, x, np.name, np.node.text))


render = render()


import sys
import re

# https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")

bname_last = 0
anchor_last = 0


from _xterm_parser import XTermParser


def more_data() -> bool:
    return False


parser = XTermParser(more_data)


def filter_in(data):
    global parser
    for event in parser.feed(data):
        clog.append(repr(event))


def filter_out(row, x, y, z):
    global clog, ansi_escape, bname_last, anchor_last
    ev = evpos[z][hash(row)]

    if row.find(":>") >= 0:
        flt = ansi_escape.sub("", row)
        len_flt = len(flt)
        while flt.find(":>") > 0:
            anchor, trail = flt.split(":>", 1)
            anchor, bname = anchor.split("<:", 1)
            if len(ev):
                posx = ev[-1][1] + anchor_last + bname_last + len(anchor)
                ev[-1][1] = posx
            else:
                posx = x

            if trail.find("<:") < 0:
                xl = x + len_flt - len(bname) - len(trail) - 4
                xr = x + len_flt - 1
                ev.append([xl, xr, bname])
                # clog.append(f"<{len(anchor)=}:{bname=} {ev=}>")
                break

            anchor_last = len(anchor)
            bname_last = len(bname) + 4

            # clog.append(f"<{len(anchor)=}:{bname=} {ev=}>")

            ev.append([posx - 1, posx, bname])
            flt = trail

        # FIXME: assuming same size buttons
        # just dividing len_flt by number of buttons

        zone = len_flt // len(ev)
        for i, pos in enumerate(ev):
            pos[0] = x + i * zone
            pos[1] = x + (i + 1) * zone - 1

    return row.replace("<:", "[ ").replace(":>", " ]")


# DECSTBM—Set Top and Bottom Margins
# https://vt100.net/docs/vt510-rm/DECSTBM.html

# DECSLRM—Set Left and Right Margins
# This control function sets the left and right margins to define the scrolling region. DECSLRM only works when vertical split screen mode (DECLRMM) is set.
# https://vt100.net/docs/vt510-rm/DECSLRM.html


class Tui:
    # use direct access, it is absolute addressing on raw terminal.
    out = out
    decvssm = False

    # save cursor
    def __enter__(self):
        # ESC("7","[?25l","[?69h")
        #        self.out("\x1b7\x1b[?25l\x1b[?69h")
        self.out("\x1b7")
        #        self.out("\x1b7\x1b[?25l")
        return self

    # restore cursor
    def __exit__(self, *tb):
        # ESC("[?69l","8","[?25h")
        #        self.out("\x1b[?69l\x1b8\x1b[?25h")
        self.out("\x1b8")
        #        self.out("\x1b8\x1b[?25h")
        pass

    def __call__(self, *a, **kw):
        global bname_last, anchor_last

        x = kw.get("x", 1)
        z = kw.get("z", 1)

        #   most term do not deal properly with vt400
        #            if decvssm:
        #                CSI(f"{x};{999}s")
        #                CSI(f"{z};{x}H\r")

        if not isinstance(a[0], str):
            import rich, io

            sio = io.StringIO()
            rich.print(*a, file=sio)
            sio.seek(0)
            block = sio.read()
        else:
            block = " ".join(a)

        # so position line by line
        filter = kw.get("filter", None)

        bname_last = 0
        anchor_last = 0

        for row in block.split("\n"):
            hr_old = damages.get(z, 0)
            hr = hash(row)
            if hr != hr_old:
                damages[z] = hr
                if filter:
                    # destroy event list ref by the old line
                    evpos.setdefault(z, {}).pop(hr_old, None)
                    evpos[z][hr] = []
                    row = filter(row, x, 0, z)

                self.out("\x1b[{};{}H{}".format(z, x, row))

            z += 1


async def taskMgr(t=1.0 / 24):
    import asyncio

    global render
    while True:
        if render.dirty:
            with render:
                for child in render.children:
                    render.draw_child(render, child, 0)
            render.dirty = False
            # flush_io() #DLE_ETX is sent for emscripten
        await asyncio.sleep(0)



from transitions import Machine
from dataclasses import dataclass
import textwrap

index = {}


class State(object):
    ...


class Draft(object):
    ...


@dataclass()
class FSM(Machine):
    states: tuple
    initial: str

    def __init__(self, *argv, **kw):
        Machine.__init__(self, states=self.states, initial=self.initial)


# TODO move the wrap out of there.


def state_refs(*argv):
    global index
    # tw = textwrap.TextWrapper(width=78)
    tw = None
    states = []
    maxchapters = len(argv)
    for chapter, component in enumerate(argv):
        source = component.script
        block = source.strip().split("\n", 1)[0]
        states.append(block)
        lines = []
        cmds = []
        for i, line in enumerate(source.split("\n")):
            if i:
                if line:
                    if line[0] != ":":
                        # markdown eats space at § begin
                        # line = line.replace('    ','\u2002\u2002')
                        if tw:
                            lines.extend(tw.wrap(line))
                        else:
                            lines.append(line)
                    else:
                        cmds.append(line[1:])
                else:
                    lines.append("")
        while len(lines) and not lines[-1]:
            lines.pop()
        lines.append("")
        lines.append("")
        lines.pop(0)
        index[block] = {
            "heading": f"Chapter {chapter+1}/{maxchapters} : {block}               ",
            "body": "\n".join(lines),
            "logic": cmds,
        }

    return tuple(states)


def build(pkg, **kw):
    global story
    steps = []

    for k, v in vars(pkg).items():
        if v in [State, Draft]:
            continue

        if isinstance(v, type):
            if issubclass(
                v,
                (
                    State,
                    Draft,
                ),
            ):
                print(k, v)
                steps.append(v)

    class Story(FSM):
        states: tuple = state_refs(*steps)
        initial: str = states[0]
        index = index

        def setup(self, pkg, **kw):
            global index
            self.name = kw.get("name", pkg.__name__)

            for state in self.states:
                store = index[state]
                store["choices"] = {}

                for cmd in store["logic"]:
                    if cmd.find(":") > 0:
                        rel, tail = cmd.split(":")
                        if rel == "?":
                            continue
                        if rel.find("->") >= 0:
                            trigger, dest = map(str.strip, rel.split("->", 1))
                            if trigger == "-":
                                trigger = dest
                            else:
                                if not trigger:
                                    trigger = dest
                                store["choices"][trigger] = tail

                            self.add_transition(trigger=trigger, source=state, dest=dest)
            return self

    story = Story()
    return story.setup(pkg, **kw)



# pygbag repl implement at the same time :
# - the python console in the style of browser javascript's one
# - the virtual gamepad
# - demo record/replay
# - event passthrough for pygame apps


HISTORY = []

from embed import set_ps1, set_ps2, prompt, readline



import sys
import aio
import platform

MODULES = list(sys.modules.keys())
MAIN = list(vars(__import__("__main__")).keys())
BUILTINS = list(vars(__import__("builtins")).keys())


def cleanup():
    import sys, random

    for mod in list(sys.modules.keys()):
        if mod in ("asyncio"):
            continue
        if not mod in MODULES:
            sys.modules.pop(mod, None)

    md = vars(__import__("__main__"))
    for name in list(md.keys()):
        if not name in MAIN:
            md.pop(name, None)

    md = vars(__import__("builtins"))
    for var in list(md.keys()):
        if not name in BUILTINS:
            md.pop(name, None)

    del mod, name, md
    __import__("importlib").invalidate_caches()
    __import__("gc").collect()

    random.seed(1)

    aio.exit = False
    aio.paused = False
    print(" - cycling done -")
    try:
        platform.set_window_title("idle")
        platform.prompt()
    except:
        pass

    def resume():
        aio.paused



STEP_INTO = False


def on():
    global STEP_INTO
    STEP_INTO = True


def off():
    global STEP_INTO
    STEP_INTO = False


def lines(frame, event, arg):
    if event != "line":
        return
    co = frame.f_code
    func_name = co.co_name
    line_no = frame.f_lineno
    filename = co.co_filename
    print(f"  {func_name} line {line_no}")


def calls(frame, event, arg):
    global STEP_INTO

    if event != "call":
        return

    co = frame.f_code
    func_name = co.co_name
    if func_name in ("write", "raw_input", "process"):
        return
    func_line_no = frame.f_lineno
    func_filename = co.co_filename

    if func_filename.startswith("/usr/lib/python3."):
        return

    if func_filename.find("/aio/") > 0:
        return

    caller = frame.f_back
    if caller:
        caller_line_no = caller.f_lineno
        caller_filename = caller.f_code.co_filename
        if caller_filename != func_filename:
            print(
                "%s() on line %s of %s from line %s of %s"
                % (
                    func_name,
                    func_line_no,
                    func_filename,
                    caller_line_no,
                    caller_filename,
                )
            )
        else:
            print(f"{func_name} {func_filename}:{caller_line_no}->{func_line_no}")

    if STEP_INTO:
        return lines
    return


"""

def trace_calls_and_returns(frame, event, arg):
    co = frame.f_code
    func_name = co.co_name
    if func_name == 'write':
        # Ignore write() calls from print statements
        return
    line_no = frame.f_lineno
    filename = co.co_filename
    if event == 'call':
        print 'Call to %s on line %s of %s' % (func_name, line_no, filename)
        return trace_calls_and_returns
    elif event == 'return':
        print '%s => %s' % (func_name, arg)
    return


"""



import sys
import builtins
import inspect


DEBUG = True
NICE = 0.010

builtins.aio = sys.modules[__name__]

# try to acquire syslog early

# TODO: use PyConfig debug flag
try:
    import embed

    def pdb(*argv):
        # print(*argv, file=sys.__stderr__)
        pass

    # builtins.pdb = embed.log
    builtins.pdb = pdb
except:

    def pdb(*argv):
        # print(*argv, file=sys.__stderr__)
        pass

    builtins.pdb = pdb


# cascade debug by default
from . import cross

if not __UPY__:
    from time import time as time_time

    # file+socket support  fopen/sopen
    from .filelike import *
else:
    import utime

    time_time = utime.ticks_ms


cross.DEBUG = DEBUG


# =========================================================================

# TODO: dbg stuff should be in the platform module in aio.cross
# usefull https://pymotw.com/3/sys/tracing.html
# upy has no trace module
if DEBUG and not __UPY__:
    import trace

    _tracer = trace.Trace(count=False, trace=True)

    def sys_trace(fn, *argv, **kw):
        global _tracer
        return _tracer.runfunc(fn, *argv, **kw)

else:

    def sys_trace(fn, *argv, **kw):
        pdb("debugging is OFF")
        return fn(*argv, **kw)


define("sys_trace", sys_trace)

try:
    import embed

    flush = embed.flush
except:

    def flush():
        sys.stdout.flush()
        sys.stderr.flush()

    if __WASM__:
        print(
            """



    WARNING: no flush method available, raw TTY may suffer



"""
        )


# ==================================================================

try:
    undefined
except:

    class sentinel:
        def __bool__(self):
            return False

        def __repr__(self):
            return "∅"

        def __nonzero__(self):
            return 0

        def __call__(self, *argv, **kw):
            if len(argv) and argv[0] is self:
                return True
            print("Null Pointer Exception")

    sentinel = sentinel()
    define("undefined", sentinel)
    del sentinel


def overloaded(i, *attrs):
    for attr in attrs:
        if attr in vars(i.__class__):
            if attr in vars(i):
                return True
    return False


define("overloaded", overloaded)
del overloaded


started = False
paused = False
exit = False
steps = []
oneshots = []
ticks = 0
protect = []
last_state = None
tasks = []
is_async_ctx = False
no_exit = True

enter = time_time()
spent = 0.00001
leave = enter + spent

from asyncio import *
from asyncio import exceptions

__run__ = run

if __UPY__:

    def _set_running_loop(l):
        pass

    sys.modules["asyncio.events"] = aio
    aio.get_running_loop = aio.get_event_loop
    events = aio
else:
    from asyncio.events import _set_running_loop


# Within a coroutine, simply use `asyncio.get_running_loop()`,
# since the coroutine wouldn't be able
# to execute in the first place without a running event loop present.
try:
    loop = get_running_loop()
except RuntimeError:
    # depending on context, you might debug or warning log that a running event loop wasn't found
    loop = get_event_loop()


# import asyncio.events
# asyncio.events._set_running_loop(loop)
_set_running_loop(loop)


sys.modules["asyncio"] = __import__(__name__)


def defer(fn, argv=(), kw={}, delay=0, framerate=60):
    global ticks, oneshots
    # FIXME: set ticks + deadline for alarm
    if __UPY__:
        tb = "n/i"
    else:
        try:
            frm = inspect.currentframe().f_back
            tb = f"{inspect.getframeinfo(frm).filename}:{frm.f_lineno}"
        except:
            tb = "no frame"

    oneshots.append(
        [
            ticks + int(delay / framerate),
            fn,
            argv,
            kw,
            tb,
        ]
    )


inloop = False


# this runs both asyncio loop and the arduino style stepper
def step(*argv):
    global inloop, last_state, paused, started, loop, oneshots, protect
    global ticks, steps, step, flush, exit, is_async_ctx
    global enter, leave, spent

    enter = time_time()

    # those hold variable that could be processed by C
    # outside the pyloop, to keep them from GC
    if protect:
        protect.clear()

    # this one is for full stop
    if not started:
        return

    try:
        # TODO: OPTIM: remove later
        if inloop:
            pdb("97: FATAL: aio loop not re-entrant !")
            paused = True
            return
        inloop = True

        if paused is not last_state:
            if not exit:
                pdb(f" - aio is {'paused' if paused else 'resuming'} -")
            else:
                print(f" - aio is exiting -")
            last_state = paused

        ticks += 1

        try:
            # defer and oneshot can run even if main loop is paused
            # eg for timekeep, or vital remote I/O sakes

            # NEED a scheduler + timesort
            early = []
            while len(oneshots):
                deferred = oneshots.pop()
                if deferred[0] > ticks:
                    deferred[0] -= 1
                    early.append(deferred)
                else:
                    _, fn, argv, kw, tb = deferred
                    try:
                        fn(*argv, **kw)
                    except Exception as e:
                        sys.print_exception(e)
                        print("--- stack -----", file=sys.__stderr__)
                        print(_, fn, argv, kw, file=sys.__stderr__)
                        print("--- stack -----", file=sys.__stderr__)
                        print("deferred from", tb, file=sys.__stderr__)

            while len(early):
                oneshots.append(early.pop())

            # TODO: fix global clock accordingly
            if not paused:
                for onestep in steps:
                    onestep()

                # loop.call_soon(loop.stop)
                # loop.run_forever()
                if started:
                    is_async_ctx = True
                    loop._run_once()
                    is_async_ctx = False

        except Exception as e:
            sys.print_exception(e)
            paused = True
            pdb("- aio paused -")

        # if using external loop handler (eg  PyOS_InputHook ) then call us again
        if started and cross.scheduler and not exit:
            cross.scheduler(step, 1)

        flush()
        inloop = False
    finally:
        leave = time_time()
        spent = leave - enter


def delta(t=None):
    global enter
    if t:
        return t - enter
    return time_time() - enter


def shed_yield():
    global enter, NICE
    return (time_time() - enter) > NICE


async def sleep_ms(ms=0):
    await sleep(float(ms) / 1000)


def _set_task_name(task, name):
    if name is not None:
        try:
            set_name = task.set_name
        except AttributeError:
            warnings.warn(
                "Task.set_name() was added in Python 3.8, "
                "the method support will be mandatory for third-party "
                "task implementations since 3.13.",
                DeprecationWarning,
                stacklevel=3,
            )
        else:
            set_name(name)


def create_task(coro, *, name=None, context=None):
    global loop, tasks

    tasks.append(coro)
    if context is None:
        # Use legacy API if context is not needed
        task = loop.create_task(coro, name=name)
    else:
        task = loop.create_task(coro, name=name, context=context)

    _set_task_name(task, name)
    return task


#
run_called = False


def is_running():
    global started
    return started


# prevent warnings in aiohttp
loop.is_running = is_running


def run(coro, *, debug=False):
    global paused, loop, started, step, DEBUG, run_called, exit

    debug = debug or DEBUG

    if coro is not None:
        wrapper = coro
        if coro.__name__ == "main":
            if "aio.fetch" in sys.modules:

                async def __main__():
                    import aio.fetch

                    await aio.fetch.preload()
                    await coro

                wrapper = __main__()
        task = loop.create_task(wrapper)
        _set_task_name(task, coro.__name__)

    if not started:
        exit = False
        run_called = True
        started = True

        # the stepper fonction when in simulator is called from pygbag.aio
        # and is blocking script
        if aio.cross.simulator:
            # cannot be sure asyncio.run() will be used
            # so don't handle loop here
            return

        #        if cross.scheduler:
        #            if debug:
        #                pdb("261: asyncio handler is", cross.scheduler)
        #            paused = False
        #            cross.scheduler(step, 1)

        # the stepper when called from  window.requestAnimationFrame()
        # https://developer.mozilla.org/en-US/docs/Web/API/window/requestAnimationFrame
        elif __EMSCRIPTEN__ or __wasi__:
            # handle special custom_site() case
            if coro.__name__ == "import_site":
                embed.run()
                run_called = False
            elif coro.__name__ == "custom_site":
                run_called = False
            elif not aio.cross.simulator:
                # let prelinker start asyncio loop
                print("AIO will auto-start at 0+, now is", embed.counter())

        # fallback to blocking asyncio
        else:
            _set_running_loop(None)
            # TODO: implement RaF from here
            try:
                loop.run_forever()
            except KeyboardInterrupt:
                print("301: closing loop")
                loop.close()

    elif run_called:
        pdb("273: aio.run called twice !!!")

    # run called after a custom_site() completion
    elif coro.__name__ != "custom_site":
        pdb("360: * custom_site done *")
    else:
        pdb("364: aio.run", coro.__name__)


def exit_now(ec):
    global exit, paused
    if exit:
        print("already exiting ...")
        return
    # rescheduling happens only where started is True
    exit = True
    while len(tasks):
        task = tasks.pop(0)
        if hasattr(task, "cancel"):
            print(f"290: canceling {task}")
            task.cancel()

    #  will prevent another asyncio loop call, we exit next cycle on oneshots queue
    paused = True
    if not __WASM__:
        pdb("291: exiting with code", ec)
    defer(__exit__, (ec,), {}, 0)


# replace sys.exit on wasm

if __WASM__:

    def __exit__(ec):
        global loop, no_exit
        if no_exit:
            pdb(f"sys.exit({ec})")
        else:
            loop.close()
        try:
            aio.recycle.cleanup()
            aio.defer(embed.prompt, (), {}, delay=300)
        except:
            pass

else:
    __exit__ = sys.exit


def aio_exit(maybecoro=0):
    if inspect.iscoroutine(maybecoro):

        async def atexit(coro):
            exit_now(await coro)

        run(atexit(maybecoro))
    else:
        # if __WASM__:
        #    pdb("309: NOT A CORO", maybecoro)
        exit_now(maybecoro)


if not __UPY__:
    sys.exit = aio_exit


# check if we have a Time handler.
try:
    Time
except:
    import time as Time


def rtclock():
    return int(Time.time() * 1_000)


class after:
    def __init__(self, oprom):
        self.oprom = oprom

    def then(self, fn, *argv, **kw):
        create_task(self.executor(fn, argv, kw))

    async def executor(self, fn, argv, kw):
        import embed

        mark = None
        value = undefined
        wit = cross.platform.window.iterator(self.oprom)
        while mark != undefined:
            value = mark
            await aio.sleep(0)
            mark = next(wit, undefined)
        del self.oprom
        if fn:
            fn(*argv, **kw)


#
import time as Time


class aioctx:
    def __init__(self, delta, coro):
        self.coro = coro
        self.tnext = Time.time() + delta
        self.tmout = 0


class _(list):
    current = None

    async def __aenter__(self):
        if self.__class__.current is None:
            self.__class__.current = aioctx(0, None)
        self.append(self.__class__.current)
        self.__class__.current = None
        if self[-1].coro is not None:
            pdb("__aenter__ awaiting", self[-1].coro)
            try:
                return await self[-1].coro
            except KeyboardInterrupt:
                aio.paused = None
                aio.loop.call_soon(aio.loop.stop)
                pdb("326: aio exit on KeyboardInterrupt")
                return await aio.sleep(0)
        else:
            print("__aenter__ no coro")
            self.__class__.current = None
            return self

    async def __aexit__(self, type, value, traceback):
        len(self) and self.pop()

    def __enter__(self):
        self.append(0)

    def __exit__(self, type, value, traceback):
        len(self) and self.pop()

    def __bool__(self):
        if self.__class__.current:
            return True
        if len(self) and self[-1]:
            return True
        return False

    def __call__(self, frametime):
        print("__call__", len(self), frametime)
        self.__class__.current = aioctx(frametime, None)
        return self

    def call(self, coro):
        print(".call", len(self), coro)
        if self.__class__.current is None:
            self.__class__.current = aioctx(0, coro)
        else:
            self.__class__.current.coro = coro
        # self.__class__.current.tmout = tmout
        return self


aio.ctx = _()



import builtins

DEBUG = False

defines = {}


def defined(plat):
    try:
        return eval(plat) or True
    except:
        return False


def define(tag, value):
    global defines, DEBUG
    if DEBUG:
        import inspect

        lf = inspect.currentframe().f_back
        fn = inspect.getframeinfo(lf).filename.rsplit("/assets/", 1)[-1]
        ln = lf.f_lineno
        info = f"{fn}:{ln}"
        defines.setdefault(tag, info)
    else:
        info = "?:?"

    redef = defined(tag)
    if redef:
        if redef is value:
            pdb(f"INFO: {tag} redefined from {defines.get(tag)} at {info}")
            pass
        else:
            pdb(
                f"""\
WARNING: {tag} was already defined
    previous {defines.get(tag)} value {redef}
    new {info} value {value}

"""
            )

    setattr(builtins, tag, value)


builtins.define = define
builtins.defined = defined



import sys
import os
import asyncio
import platform
import json
import warnings
import shutil

from pathlib import Path
from urllib.parse import urlencode

preload_list = []
preloaded = []

if 0:
    FS_SILENT = True
    FS_DEBUG = False
else:
    FS_SILENT = False
    FS_DEBUG = True


def FS(tree, silent=FS_SILENT, debug=FS_DEBUG):
    global preload_list

    target = [""]
    path = [""]

    base_url = ""
    base_path = ""

    last = 0
    trail = False

    def make_src_dst(base_url, base_path, path, target):
        global preload_list
        nonlocal debug
        dst = target.copy()
        dst.extend(path[1:])

        preload_list.append([base_path + "/".join(path), "/".join(dst)])
        if debug:
            print(preload_list[-1], "write", current, last)

    for l in map(str.rstrip, tree.split("\n")):
        if not silent:
            print("HTMLFS:", base_url, base_path, "=>", target)

        # skip blanks
        if not l:
            continue

        # skip root of tree, use current url as base_url
        if l == ".":
            preload_list.append(["", "."])
            continue

        if l.find(" ~") > 0:
            # reset source
            path = [""]

            base_path, trail = map(str.strip, l.split(" ~", 1))

            base_path = base_path.strip("/")

            trail = trail.strip("/")

            # set destination
            target = [trail]
            if debug:
                print(f"{base_path=}, {target=}")
            continue

        if l.startswith("http"):
            # found a base url
            base_url = l
            if base_url.find("/github.com/") > 0:
                base_url = base_url.replace("/github.com/", "/raw.githubusercontent.com/", 1)

                if base_url.find("/tree/") > 0:
                    base_url = base_url.replace("/tree/", "/", 1)

            base_url = base_url.rstrip("/") + "/"

            preload_list.append([base_url, "."])
            continue

        if l.find(" ") < 0:
            continue

        pos, elem = l.rsplit(" ", 1)
        current = (1 + len(pos)) // 4

        if not silent:
            print(l[4:])

        if current <= last:
            make_src_dst(base_url, base_path, path, target)
            while len(path) > current:
                path.pop()
        else:
            trail = True

        if debug:
            print(f"91: {pos=} {elem=} {current=} {path=} {last=}")

        if len(path) < current + 1:
            path.append(elem)
        path[current] = elem
        last = current

    if trail:
        make_src_dst(base_url, base_path, path, target)
    print("-" * 80)
    for x in preload_list:
        print(x)
    print("-" * 80)
    return preload_list


async def preload(chroot=None, chdir=True, silent=FS_SILENT, debug=FS_DEBUG, isolated=False):
    global preload_list, preloaded

    # if not using FS, do not change directory
    if not len(preload_list):
        if isolated:
            return []
        # return all previous filesets
        return preloaded.copy()

    cwd = Path.cwd()

    # if using FS, always go to tempdir
    if chroot is None:
        chroot = Path(__import__("tempfile").gettempdir())

    base_url = ""
    fileset = []

    while len(preload_list):
        url, strfilename = preload_list.pop(0)

        if strfilename == ".":
            base_url = url
            if debug:
                print(f"{base_url=} set")
            continue

        filename = chroot / strfilename

        full_url = base_url + url

        if debug:
            print(f"153: REMOTE {full_url} LOCAL: {filename}")

        if not filename.is_file():
            filename.parent.mkdir(parents=True, exist_ok=True)
            if sys.platform in ("emscripten", "wasi"):
                async with platform.fopen(full_url, "rb") as source:
                    with open(filename, "wb") as destination:
                        destination.write(source.read())
            else:
                localfile = Path(cwd.as_posix() + "/" + url)
                if localfile.is_file():
                    if debug:
                        print(f"GET {localfile=} {chroot=} {filename=}")
                    shutil.copy(localfile, filename)
                else:
                    warnings.warn(f"{localfile=} not found")

        fileset.append(filename)

        if not silent:
            print("FS:174:", filename)

    if not isolated:
        preloaded.extend(fileset)

    if chdir:
        os.chdir(chroot)

    if isolated:
        return fileset

    await asyncio.sleep(0)

    if debug:
        print("184: preloaded count=", len(preloaded))
    return preloaded.copy()


class RequestHandler:
    """
    WASM compatible request handler
    auto-detects emscripten environment and sends requests using JavaScript Fetch API
    """

    GET = "GET"
    POST = "POST"
    _js_code = ""
    _init = False

    def __init__(self):
        self.is_emscripten = sys.platform == "emscripten"
        if not self._init:
            self.init()
        self.debug = True
        self.result = None
        if not self.is_emscripten:
            try:
                import requests

                self.requests = requests
            except ImportError:
                pass

    def init(self):
        if self.is_emscripten:
            self._js_code = """
window.Fetch = {}
// generator functions for async fetch API
// script is meant to be run at runtime in an emscripten environment
// Fetch API allows data to be posted along with a POST request
window.Fetch.POST = function * POST (url, data)
{
    // post info about the request
    console.log('POST: ' + url + 'Data: ' + data);
    var request = new Request(url, {headers: {'Accept': 'application/json','Content-Type': 'application/json'},
        method: 'POST',
        body: data});
    var content = 'undefined';
    fetch(request)
   .then(resp => resp.text())
   .then((resp) => {
        console.log(resp);
        content = resp;
   })
   .catch(err => {
         // handle errors
         console.log("An Error Occurred:")
         console.log(err);
    });
    while(content == 'undefined'){
        yield;
    }
    yield content;
}
// Only URL to be passed
// when called from python code, use urllib.parse.urlencode to get the query string
window.Fetch.GET = function * GET (url)
{
    console.log('GET: ' + url);
    var request = new Request(url, { method: 'GET' })
    var content = 'undefined';
    fetch(request)
   .then(resp => resp.text())
   .then((resp) => {
        console.log(resp);
        content = resp;
   })
   .catch(err => {
         // handle errors
         console.log("An Error Occurred:");
         console.log(err);
    });
    while(content == 'undefined'){
        // generator
        yield;
    }

    yield content;
}
            """
            try:
                platform.window.eval(self._js_code)
            except AttributeError:
                self.is_emscripten = False

    @staticmethod
    def read_file(file):
        # synchronous reading of file intended for evaluating on initialization
        # use async functions during runtime
        with open(file, "r") as f:
            data = f.read()
        return data

    @staticmethod
    def print(*args, default=True):
        try:
            for i in args:
                platform.window.console.log(i)
        except AttributeError:
            pass
        except Exception as e:
            return e
        if default:
            print(*args)

    async def get(self, url, params=None, doseq=False):
        # await asyncio.sleep(5)
        if params is None:
            params = {}
        if self.is_emscripten:
            query_string = urlencode(params, doseq=doseq)
            await asyncio.sleep(0)
            content = await platform.jsiter(platform.window.Fetch.GET(url + "?" + query_string))
            if self.debug:
                self.print(content)
            self.result = content
        else:
            self.result = self.requests.get(url, params).text
        return self.result

    # def get(self, url, params=None, doseq=False):
    #     return await self._get(url, params, doseq)

    async def post(self, url, data=None):
        if data is None:
            data = {}
        if self.is_emscripten:
            await asyncio.sleep(0)
            content = await platform.jsiter(platform.window.Fetch.POST(url, json.dumps(data)))
            if self.debug:
                self.print(content)
            self.result = content
        else:
            self.result = self.requests.post(
                url, data, headers={"Accept": "application/json", "Content-Type": "application/json"}
            ).text
        return self.result

    # def post(self, url, data=None):
    #     return await self._post(url, data)



from pathlib import Path

dbg = False

# q:what to do with the extreme case $HOME/main.py ?
# or folders > 512MiB total
# a: maybe break on too many files around the yield


IGNORE = """
/.mypy_cache
/.ssh
/.local
/.config
/.git
/.github
/.vscode
/.idea
/dist
/build
/venv
/ignore
/ATTIC
""".strip().split(
    "\n"
)

SKIP_EXT = ["pyc", "pyx", "pyd", "pyi", "exe", "log", "DS_Store"]


def filter(walked):
    global dbg, IGNORE, SKIP_EXT
    for folder, filenames in walked:
        blocking = False

        for block in IGNORE:
            if not block:
                continue
            if folder.match(block):
                if dbg:
                    print("REJ 1", folder)
                blocking = True
                break

            fx = folder.as_posix()

            if fx.startswith(f"{block}/"):
                if dbg:
                    print("REJ 2", folder)
                blocking = True
                break

        if blocking:
            continue

        for filename in filenames:
            if filename in [".gitignore"]:
                if dbg:
                    print("REJ 3", folder, filename)
                continue

            ext = filename.rsplit(".", 1)[-1].lower()
            if ext in SKIP_EXT:
                if dbg:
                    print("REJ 4", folder, filename)
                continue

            yield Path(folder), Path(folder).joinpath(filename)



import sys
import os

# this aio is the support/cross one, not the current file.
import aio

# to allow "import pygbag.aio as asyncio"
sys.modules["pygbag.aio"] = aio

if hasattr(os, "uname") and not os.uname().machine.startswith("wasm"):
    import time
    from pathlib import Path

    try:
        # import aioconsole, aiohttp
        import aiohttp
    except Exception as e:
        print(
            e,
            f"""

pygbag simulator rely on : aiohttp asyncio_socks_server token_util
please use :

    {sys.executable} -m pip install aiohttp asyncio_socks_server token_util

""",
        )
        raise SystemExit

    import aio.pep0723

    if aio.pep0723.Config.dev_mode:
        aio.pep0723.Config.PKG_INDEXES.extend(["http://localhost:8000/archives/repo/"])
    else:
        aio.pep0723.Config.PKG_INDEXES.extend(["https://pygame-web.github.io/archives/repo/"])

    import pygbag.__main__

    #
    #    async def custom_async_input():
    #        import platform
    #        if platform.window.RAW_MODE:
    ## TODO: FIXME: implement embed.os_read and debug focus handler
    #            #return await asyncio.sleep(0)
    #            ...
    #        return await aioconsole.ainput("››› ")

    aio.loop.create_task(
        pygbag.__main__.import_site(
            sourcefile=sys.argv[0],
            simulator=True,
            #            async_input=custom_async_input,
            async_input=None,
            async_pkg=aio.pep0723.check_list(filename=sys.argv[0]),
        )
    )

    # todo use a thread for watchdog.

    # simulate .requestAnimationFrame() infinite loop

    # make aiohttp happy
    aio.started = True

    while not aio.exit:
        next = time.time() + 0.0155
        try:
            aio.loop._run_once()
        except KeyboardInterrupt:
            print("45: KeyboardInterrupt")

        dt = next - time.time()
        if dt < 0:
            past = int(-dt * 1000)
            # do not spam for <4 ms late (50Hz vs 60Hz)
            if past > 4:
                print(f"\raio: violation frame is {past} ms late")
            # too late do not sleep at all
        else:
            time.sleep(dt)
    print("sim loop exited")
    sys.stdout.flush()



import sys
import pygbag
from pathlib import Path

## ====================== pygame

# def pg_load(fn, alpha=True):
#    import pygame
#    from pathlib import Path

#    if Path(fn).is_file():
#        media = pygame.image.load(fn)
#    else:
#        media = pygame.image.load( Path(__file__).parent / "offline.png" )

#    try:
#        if alpha:
#            return media.convert_alpha()
#        else:
#            return media.convert()
#    # offscreen case
#    except:
#        return media


# ======================= network


import pygbag.aio as asyncio

if aio.cross.simulator:
    sys.path.append(str(Path(aio.__file__).parent.parent.parent))

print("   ================ PygBag utilities loaded ================== ")


def host():
    global proxy
    print("serving sockv5 ...")

    from asyncio_socks_server.values import SocksAuthMethod
    from asyncio_socks_server.config import Config
    from asyncio_socks_server.proxyman import ProxyMan

    cfg = Config()

    cfg.update_config(
        {
            "LISTEN_HOST": "127.0.0.1",
            "LISTEN_PORT": 8001,
            "AUTH_METHOD": SocksAuthMethod.NO_AUTH,
            "ACCESS_LOG": False,
            "STRICT": False,
            "DEBUG": False,
            "USERS": {},
        }
    )

    proxy = ProxyMan(cfg)
    aio.create_task(proxy.start_server())


host()


async def connect():
    import aiohttp
    from aiohttp_socks import ProxyType, ProxyConnector, ChainProxyConnector

    async def fetch(url):
        connector = ProxyConnector.from_url("socks5://user:password@127.0.0.1:8001")
        async with aiohttp.ClientSession(connector=connector) as session:
            async with session.get(url) as response:
                return await response.text()

    print(await fetch("https://example.com/"))


# await connect()

if __name__ == "__main__":

    async def main():
        asyncio.run(main())



#
# https://peps.python.org/pep-0722/ – Dependency specification for single-file scripts
# https://peps.python.org/pep-0723
# https://peps.python.org/pep-0508/ – Dependency specification for Python Software Packages
# https://setuptools.pypa.io/en/latest/userguide/ext_modules.html
#

import sys
import os
from pathlib import Path
import glob

import re

print(sys.path)
import tomllib

import json

import importlib
import installer
import pyparsing
from packaging.requirements import Requirement

from aio.filelike import fopen

import platform
import platform_wasm.todo


# TODO: maybe control wheel cache with $XDG_CACHE_HOME/pip


# store installed wheel somewhere
env = Path(os.getcwd()) / "build" / "env"
env.mkdir(parents=True, exist_ok=True)

# we want host to load wasm packages too
# so make pure/bin folder first for imports

if env.as_posix() not in sys.path:
    sys.path.insert(0, env.as_posix())

sconf = __import__("sysconfig").get_paths()
sconf["purelib"] = sconf["platlib"] = env.as_posix()

if sconf["platlib"] not in sys.path:
    sys.path.append(sconf["platlib"])

PATCHLIST = []
HISTORY = []


class Config:
    READ_722 = False
    READ_723 = True
    BLOCK_RE_722 = r"(?i)^#\s+script\s+dependencies:\s*$"
    BLOCK_RE_723 = r"(?m)^# /// (?P<type>[a-zA-Z0-9-]+)$\s(?P<content>(^#(| .*)$\s)+)^# ///$"
    PKG_INDEXES = []
    REPO_INDEX = "index.json"
    REPO_DATA = "repodata.json"
    repos = []
    pkg_repolist = []
    dev_mode = ".-X.dev." in ".".join([""] + sys.orig_argv + [""])

    mapping = {
        "pygame": "pygame.base",
        "pygame_ce": "pygame.base",
        "python_i18n": "i18n",
        "pillow": "PIL",
    }


def read_dependency_block_722(code):
    # Skip lines until we reach a dependency block (OR EOF).
    has_block = False
    # Read dependency lines until we hit a line that doesn't
    # start with #, or we are at EOF.
    for line in code.split("\n"):
        if not has_block:
            if re.match(Config.BLOCK_RE_722, line):
                has_block = True
            continue

        if not line.startswith("#"):
            break
        # Remove comments. An inline comment is introduced by
        # a hash, which must be preceded and followed by a
        # space.
        line = line[1:].split(" # ", maxsplit=1)[0]
        line = line.strip()
        # Ignore empty lines
        if not line:
            continue
        # Try to convert to a requirement. This will raise
        # an error if the line is not a PEP 508 requirement
        yield Requirement(line)


def read_dependency_block_723(code):
    # Skip lines until we reach a dependency block (OR EOF).
    has_block = False

    content = []
    for line in code.split("\n"):
        if not has_block:
            if line.strip() == "# /// pyproject":
                has_block = True
            continue

        if not line.startswith("#"):
            break

        if line.strip() == "# ///":
            break

        content.append(line[2:])
    struct = tomllib.loads("\n".join(content))

    print(json.dumps(struct, sort_keys=True, indent=4))

    project = struct.get("project", {"dependencies": []})
    for dep in project.get("dependencies", []):
        yield dep


def read_dependency_block_723x(script):
    name = "pyproject"
    matches = list(filter(lambda m: m.group("type") == name, re.finditer(Config.BLOCK_RE_723, script)))
    if len(matches) > 1:
        raise ValueError(f"Multiple {name} blocks found")
    elif len(matches) == 1:
        print(tomllib.loads(matches[0]))
        yield "none"
    else:
        return None


def install(pkg_file, sconf=None):
    global HISTORY
    from installer import install
    from installer.destinations import SchemeDictionaryDestination
    from installer.sources import WheelFile

    # Handler for installation directories and writing into them.
    destination = SchemeDictionaryDestination(
        sconf or __import_("sysconfig").get_paths(),
        interpreter=sys.executable,
        script_kind="posix",
    )

    try:
        with WheelFile.open(pkg_file) as source:
            install(
                source=source,
                destination=destination,
                # Additional metadata that is generated by the installation tool.
                additional_metadata={
                    "INSTALLER": b"pygbag",
                },
            )
        if pkg_file not in HISTORY:
            HISTORY.append(pkg_file)
            importlib.invalidate_caches()
        print(f"142: {pkg_file} installed")
    except FileExistsError as ex:
        print(f"38: {pkg_file} already installed (or partially)", ex)
    except Exception as ex:
        pdb(f"82: cannot install {pkg_file}")
        sys.print_exception(ex)


async def async_imports_init():
    ...


#    see pythonrc
#            if not len(Config.repos):
#                for cdn in (Config.PKG_INDEXES or PyConfig.pkg_indexes):
#                    async with platform.fopen(Path(cdn) / Config.REPO_DATA) as source:
#                        Config.repos.append(json.loads(source.read()))
#
#                DBG("1203: FIXME (this is pyodide maintened stuff, use PEP723 asap) referenced packages :", len(cls.repos[0]["packages"]))
#


async def async_repos():
    abitag = f"cp{sys.version_info.major}{sys.version_info.minor}"

    apitag = __import__("sysconfig").get_config_var("HOST_GNU_TYPE")
    apitag = apitag.replace("-", "_")
    print("163: async_repos", Config.PKG_INDEXES)
    for repo in Config.PKG_INDEXES:
        if apitag.find("mvp") > 0:
            idx = f"{repo}index.json"
        else:
            idx = f"{repo}index-bi.json"
        async with fopen(idx, "r", encoding="UTF-8") as index:
            try:
                data = index.read()
                if isinstance(data, bytes):
                    data = data.decode()
                data = data.replace("<abi>", abitag)
                data = data.replace("<api>", apitag)
                repo = json.loads(data)
            except:
                pdb(f"110: {repo=}: malformed json index {data}")
                continue
            if repo not in Config.pkg_repolist:
                Config.pkg_repolist.append(repo)

    repo = None
    if Config.dev_mode > 0:
        for idx, repo in enumerate(Config.pkg_repolist):
            try:
                repo["-CDN-"] = Config.PKG_INDEXES[idx]
            except Exception as e:
                sys.print_exception(e)

    if not aio.cross.simulator:
        import platform

        print("193:", platform.window.location.href)
        if platform.window.location.href.startswith("http://localhost:8"):
            for idx, repo in enumerate(Config.pkg_repolist):
                repo["-CDN-"] = "http://localhost:8000/archives/repo/"
        elif platform.window.location.href.startswith("https://pmp-p.ddns.net/pygbag"):
            for idx, repo in enumerate(Config.pkg_repolist):
                repo["-CDN-"] = "https://pmp-p.ddns.net/archives/repo/"
        elif platform.window.location.href.startswith("http://192.168.1.66/pygbag"):
            for idx, repo in enumerate(Config.pkg_repolist):
                repo["-CDN-"] = "http://192.168.1.66/archives/repo/"
    if repo:
        print(
            f"""

===============  REDIRECTION TO DEV HOST {repo['-CDN-']}  ================
{abitag=}
{apitag=}

"""
        )


async def install_pkg(sysconf, wheel_url, wheel_pkg):
    target_filename = f"/tmp/{wheel_pkg}"
    async with fopen(wheel_url, "rb") as pkg:
        with open(target_filename, "wb") as target:
            target.write(pkg.read())
    install(target_filename, sysconf)


def do_patches():
    global PATCHLIST
    # apply any patches
    while len(PATCHLIST):
        dep = PATCHLIST.pop(0)
        print(f"254: patching {dep}")
        try:
            import platform

            platform.patches.pop(dep)()
        except Exception as e:
            sys.print_exception(e)


# FIXME: HISTORY and invalidate caches
async def pip_install(pkg, sysconf={}):
    global sconf
    if pkg in HISTORY:
        return

    print("253: searching", pkg)

    if not sysconf:
        sysconf = sconf

    wheel_url = ""

    # hack for WASM wheel repo
    if pkg.lower() in Config.mapping:
        pkg = Config.mapping[pkg.lower()]
        if pkg in HISTORY:
            return
        print("279: package renamed to", pkg)

    if pkg in platform.patches:
        if not pkg in PATCHLIST:
            PATCHLIST.append(pkg)

    for repo in Config.pkg_repolist:
        if pkg in repo:
            wheel_url = f"{repo['-CDN-']}{repo[pkg]}#"

    # try to get a pure python wheel from pypi
    if not wheel_url:
        try:
            async with fopen(f"https://pypi.org/simple/{pkg}/") as html:
                if html:
                    for line in html.readlines():
                        if line.find("href=") > 0:
                            if line.find("-py3-none-any.whl") > 0:
                                wheel_url = line.split('"', 2)[1]
                else:
                    print("283: ERROR: cannot find package :", pkg)
        except FileNotFoundError:
            print("285: ERROR: cannot find package :", pkg)
            return

        except:
            print("289: ERROR: cannot find package :", pkg)
            return

    if wheel_url:
        try:
            wheel_pkg, wheel_hash = wheel_url.rsplit("/", 1)[-1].split("#", 1)
            await install_pkg(sysconf, wheel_url, wheel_pkg)
            if pkg not in HISTORY:
                HISTORY.append(pkg)
        except:
            print("299: INVALID", pkg, "from", wheel_url)


PYGAME = 0


async def parse_code(code, env):
    global PATCHLIST, PYGAME

    maybe_missing = []

    if Config.READ_722:
        for req in read_dependency_block_722(code):
            pkg = str(req)
            if (env / pkg).is_dir():
                print("found in env :", pkg)
                continue
            elif pkg not in maybe_missing:
                # do not change case ( eg PIL )
                maybe_missing.append(pkg.lower().replace("-", "_"))

    if Config.READ_723:
        for req in read_dependency_block_723(code):
            pkg = str(req)
            if (env / pkg).is_dir():
                print("found in env :", pkg)
                continue
            elif pkg not in maybe_missing:
                # do not change case ( eg PIL )
                maybe_missing.append(pkg.lower().replace("-", "_"))

    still_missing = []

    import platform

    for dep in maybe_missing:
        if dep in platform.patches:
            PATCHLIST.append(dep)

        # special case of pygame code in pygbag site-packages
        if dep == "pygame.base" and not PYGAME:
            PYGAME = 1
            still_missing.append(dep)
            continue

        if not importlib.util.find_spec(dep) and dep not in still_missing:
            still_missing.append(dep)
        else:
            print("found in path :", dep)

    return still_missing


# parse_code does the patching
# this is not called by pythonrc
async def check_list(code=None, filename=None):
    global PATCHLIST, async_imports_init, async_repos, env, sconf
    print()
    print("-" * 11, "computing required packages", "-" * 10)

    # pythonrc is calling aio.pep0723.parse_code not check_list
    # so do patching here
    patchlevel = platform_wasm.todo.patch()
    if patchlevel:
        print("264:parse_code() patches loaded :", list(patchlevel.keys()))
        platform_wasm.todo.patch = lambda: None
        # and only do that once and for all.
        await async_imports_init()
        await async_repos()
        del async_imports_init, async_repos

    # mandatory
    importlib.invalidate_caches()

    if code is None:
        code = open(filename, "r").read()

    still_missing = await parse_code(code, env)

    # is there something to do ?
    if len(still_missing):
        importlib.invalidate_caches()

        # TODO: check for possible upgrade of env/* pkg

        maybe_missing = still_missing
        still_missing = []

        for pkg in maybe_missing:
            hit = ""
            for repo in Config.pkg_repolist:
                wheel_pkg = repo.get(pkg, "")
                if wheel_pkg:
                    wheel_url = repo["-CDN-"] + "/" + wheel_pkg
                    wheel_pkg = wheel_url.rsplit("/", 1)[-1]
                    await install_pkg(sconf, wheel_url, wheel_pkg)
                    hit = pkg

            if len(hit):
                print("found on pygbag repo and installed to env :", hit)
            else:
                still_missing.append(pkg)

        for pkg in still_missing:
            di = f"{(env / pkg).as_posix()}-*.dist-info"
            gg = glob.glob(di)
            if gg:
                print("found in env :", gg[0].rsplit("/", 1)[-1])
                continue

            pkg_final = pkg.replace("-", "_")
            if (env / pkg_final).is_dir():
                print("found in env :", pkg)
                continue
            await pip_install(pkg_final, sconf)

    # wasm compilation
    if not aio.cross.simulator:
        import platform
        import asyncio

        platform.explore(sconf["platlib"])
        await asyncio.sleep(0)

    do_patches()

    print("-" * 40)
    print()

    return still_missing


# aio.pep0723



import sys
import os
import aio

# https://bugs.python.org/issue34616
# https://github.com/ipython/ipython/blob/320d21bf56804541b27deb488871e488eb96929f/IPython/core/interactiveshell.py#L121-L150

import asyncio
import ast
import types
import inspect

if not __UPY__:
    try:
        from . import repl
    except:
        repl = None

    import code

    class AsyncInteractiveConsole(code.InteractiveConsole):
        instance = None
        console = None
        # TODO: use PyConfig interactive flag
        muted = True

        def __init__(self, locals, **kw):
            super().__init__(locals)
            self.compile.compiler.flags |= ast.PyCF_ALLOW_TOP_LEVEL_AWAIT
            self.line = ""
            self.buffer = []
            self.one_liner = True
            self.opts = kw
            self.shell = self.opts.get("shell", None)

            if self.shell is None:

                class shell:
                    coro = []
                    is_interactive = None

                    @classmethod
                    def parse_sync(shell, line, **env):
                        print("NoOp shell", line)

                self.shell = shell

            self.rv = None

            try:
                sys.ps1
            except AttributeError:
                sys.ps1 = ">A> "

            try:
                sys.ps2
            except AttributeError:
                sys.ps2 = "--- "

        def runsource(self, source, filename="<stdin>", symbol="single"):
            if len(self.buffer) > 1:
                symbol = "exec"

            try:
                code = self.compile(source, filename, symbol)
            except SyntaxError:
                if self.one_liner:
                    if self.shell.parse_sync(self.line):
                        return
                self.showsyntaxerror(filename)
                return False

            except (OverflowError, ValueError):
                # Case 1
                self.showsyntaxerror(filename)
                return False

            if code is None:
                # Case 2
                return True

            # Case 3
            self.runcode(code)
            return False

        def runcode(self, code):
            if repl:
                repl.set_ps1()
            self.rv = undefined

            bc = types.FunctionType(code, self.locals)
            try:
                self.rv = bc()
            except SystemExit:
                aio.exit_now(0)

            except KeyboardInterrupt as ex:
                print(ex, file=sys.__stderr__)
                raise

            except ModuleNotFoundError as ex:
                want = str(ex).split("'")[1]
                print("189 : FIXME sync->async->sync import bytecode retry in non interactive")
                print(f'await aio.pep0723.pip_install("{want}");import {want}')

                async def import_now():
                    nonlocal want
                    await aio.pep0723.pip_install(want)
                    vars(__import__("__main__"))[want] = __import__(want)

                self.shell.coro.append(import_now())

            except BaseException as ex:
                if self.one_liner:
                    shell = self.opts.get("shell", None)
                    if shell:
                        # coro maybe be filled by shell exec
                        if shell.parse_sync(self.line):
                            return
                sys.print_exception(ex, limit=-1)

            finally:
                self.one_liner = True

        def banner(self):
            if self.muted:
                return
            cprt = 'Type "help", "copyright", "credits" or "license" for more information.'

            self.write("\nPython %s on %s\n%s\n" % (sys.version, sys.platform, cprt))

        def prompt(self, prompt=None):
            if not self.__class__.muted and self.shell.is_interactive:
                import platform

                # that is the browser one
                # platform.prompt(prompt or sys.ps1)
                if repl:
                    repl.prompt()

        async def input_console(self, prompt=">I> "):
            if len(self.buffer):
                return self.buffer.pop(0)

            # if program wants I/O do not empty buffers
            if self.shell.is_interactive:
                if repl and not aio.cross.simulator:
                    maybe = repl.readline()
                elif aio.async_input:
                    maybe = await aio.async_input()
                else:
                    maybe = ""

                if len(maybe):
                    return maybe
            return None

        # can be used to mix console and app
        async def interact_step(self, prompt=None):
            if aio.exit:
                return

            if prompt is None:
                prompt = sys.ps1

            try:
                try:
                    self.line = await self.input_console(prompt)
                    if self.line is None:
                        return

                except EOFError:
                    self.write("\n")
                    return
                else:
                    if self.push(self.line):
                        if self.one_liner:
                            prompt = sys.ps2
                            if repl:
                                repl.set_ps2()
                            print("Sorry, multi line input editing is not supported", file=sys.stderr)
                            self.one_liner = False
                            self.resetbuffer()
                        else:
                            return
                    else:
                        prompt = sys.ps1

            except KeyboardInterrupt:
                self.write("\nKeyboardInterrupt\n")
                self.resetbuffer()
                self.one_liner = True

            if aio.exit:
                return

            try:
                # if async prepare is required
                while len(self.shell.coro):
                    self.rv = await self.shell.coro.pop(0)

                # if self.rv not in [undefined, None, False, True]:
                if inspect.isawaitable(self.rv):
                    await self.rv
            except RuntimeError as re:
                if str(re).endswith("awaited coroutine"):
                    ...
                else:
                    sys.print_exception(ex)

            except Exception as ex:
                print(type(self.rv), self.rv)
                sys.print_exception(ex)

            self.prompt()

        async def interact(self):
            # in repl+raw mode we don't want that loop to read input
            import sys
            from platform import window

            while not aio.exit:
                await asyncio.sleep(0)
                if window.RAW_MODE:
                    continue
                await self.interact_step(sys.ps1)
            aio.exit_now(0)

        @classmethod
        def make_instance(cls, shell, ns="__main__"):
            cls.instance = cls(
                vars(__import__(ns)),
                shell=shell,
            )
            import platform

            platform.shell = shell
            shell.runner = cls.instance
            del AsyncInteractiveConsole.make_instance

        @classmethod
        def start_console(cls, shell, ns="__main__"):
            """will only start a console, not async import system"""
            if cls.instance is None:
                cls.make_instance(shell, ns)

            if cls.console is None:
                asyncio.create_task(cls.instance.interact())
                cls.console = cls.instance

        @classmethod
        async def start_toplevel(cls, shell, console=True, ns="__main__"):
            """start async import system with optionnal async console"""
            if cls.instance is None:
                cls.make_instance(shell, ns)
                # await cls.instance.async_repos()

            if console:
                cls.start_console(shell, ns=ns)

else:
    # TODO upy event driven async repl

    class AsyncInteractiveConsole:
        ...



# builtins.__EMSCRIPTEN__
# builtins.__WASI__
# builtins.__ANDROID__   also defines __ANDROID_API__
# builtins.__UPY__ too can point to this module

import sys, os, builtins
import json

builtins.builtins = builtins

builtins.true = True
builtins.false = False


def breakpointhook(*argv, **kw):
    aio.paused = True


def shed_yield():
    # TODO: coroutine id as pid
    print("21", time_time() - aio.enter, aio.spent)
    return True


this = __import__(__name__)

# those  __dunder__ are usually the same used in C conventions.

# try:
#    __UPY__
# except:
#    if hasattr(sys.implementation, "_mpy"):
#        builtins.__UPY__ = this
#    else:
#        builtins.__UPY__ = None


# force use a fixed, tested version of uasyncio to avoid non-determinism
if __UPY__:
    sys.modules["sys"] = sys
    sys.modules["builtins"] = builtins

#    try:
#        from . import uasyncio as uasyncio
#
#        print("Warning : using WAPY uasyncio")
#    except Exception as e:
#        sys.print_exception(e)

else:
    sys.breakpointhook = breakpointhook

    # fallback to asyncio based implementation
    try:
        from . import uasyncio_cpy as uasyncio
    except:
        pdb("INFO: no uasyncio implementation found")
        uasyncio = aio

    sys.modules["uasyncio"] = uasyncio


# detect if cpython is really running on a emscripten browser

if hasattr(sys, "_emscripten_info"):
    is_browser = not sys._emscripten_info.runtime.startswith("Node.js")
    builtins.__EMSCRIPTEN__ = this
    try:
        from embed import *
        from platform import *
        from embed_emscripten import *
        from embed_browser import window, document, navigator
        from embed_browser import Audio, File, Object
        from embed_browser import fetch, console, prompt, alert, confirm

        # broad pyodide compat
        sys.modules["js"] = this  # instead of just sys.modules["embed_browser"]

        Object_type = type(Object())
    except Exception as e:
        sys.print_exception(e)
        pdb(__file__, ":47 no browser/emscripten modules yet", e)

    def ffi(arg):
        return window.JSON.parse(json.dumps(arg))

    async def jsiter(iterator):
        mark = None
        value = undefined
        while mark != undefined:
            value = mark
            await aio.sleep(0)
            mark = next(iterator, undefined)
        return value

    async def jsprom(prom):
        return await jsiter(platform.window.iterator(prom))

else:
    is_browser = False
    pdb("101: no emscripten browser interface")
    builtins.__EMSCRIPTEN__ = None


def init_platform(embed):
    # simulator won't run javascript for now
    if not hasattr(embed, "run_script"):
        pdb("186: no js engine")
        return False

    import json

    def js(code, delay=0):
        # keep a ref till next loop
        aio.protect.append(code)
        if not delay:
            result = embed.run_script(f"JSON.stringify({code})")
            if result is not None:
                return json.loads(result)
        elif delay < 0:
            embed.run_script(code)
        else:
            embed.run_script(code, int(delay))

    embed.js = js

    if __WASM__:
        import _thread

        try:
            _thread.start_new_thread(lambda: None, ())
        except RuntimeError:
            pdb("WARNING: that wasm build does not support threads")


# ========================================== DOM EVENTS ===============

if is_browser:
    # implement "js.new"

    def new(oclass, *argv):
        from embed_browser import Reflect, Array

        return Reflect.construct(oclass, Array(*argv))

    # dom events
    class EventTarget:
        clients = {}
        events = []

        def addEventListener(self, host, type, listener, options=None, useCapture=None):
            cli = self.__class__.clients.setdefault(type, [])
            cli.append(listener)

        def build(self, evt_name, jsondata):
            try:
                self.__class__.events.append([evt_name, json.loads(jsondata.strip('"'))])
            except Exception as e:
                sys.print_exception(e)
                print(jsondata)

        # def dispatchEvent
        async def rpc(self, method, *argv):
            import inspect

            # TODO resolve whole path
            if hasattr(__import__("__main__"), method):
                client = getattr(__import__("__main__"), method)
                is_coro = inspect.iscoroutinefunction(client)
                if is_coro:
                    await client(*argv)
                else:
                    client(*argv)
            else:
                print(f"RPC not found: {method}{argv}")

        # This is a green thread handling events from js -> python

        async def process(self):
            import inspect
            from types import SimpleNamespace

            # notify event queue we are ready to process
            window.python.is_ready = 1

            while not aio.exit:
                if len(self.events):
                    evtype, evdata = self.events.pop(0)
                    if evtype == "rpc":
                        print("rpc.id", evdata.pop("rpcid"))
                        await self.rpc(evdata.pop("call"), *evdata.pop("argv"))
                        continue

                    discarded = True
                    for client in self.clients.get(evtype, []):
                        is_coro = inspect.iscoroutinefunction(client)
                        discarded = False
                        sns = SimpleNamespace(**evdata)
                        if is_coro:
                            await client(sns)
                        else:
                            client(sns)

                    if discarded:
                        console.log(f"221: DISCARD : {evtype} {evdata}")

                await aio.sleep(0)

    EventTarget = EventTarget()
    add_event_listener = EventTarget.addEventListener

# =============================  PRELOADING      ==============================

preloading = -1
prelist = {}
ROOTDIR = f"/data/data/{sys.argv[0]}/assets"


def explore(root):
    global prelist, preloading

    if preloading < 0:
        preloading = 0

    import shutil

    for current, dirnames, filenames in os.walk(root):
        for filename in filenames:
            if filename.endswith(".so"):
                preloading += 1
                src = f"{current}/{filename}"
                embed.preload(src)


def fix_preload_table():
    global prelist, preloadedWasm, preloadedImages, preloadedAudios

    if embed.counter() < 0:
        pdb("212: asset manager not ready 0>", embed.counter())
        aio.defer(fix_preload_table, (), {}, delay=60)

    for (
        src,
        dst,
    ) in prelist.items():
        ext = dst.rsplit(".", 1)[-1]
        if ext in preloadedImages:
            ptype = "preloadedImages"
        elif ext in preloadedAudios:
            ptype = "preloadedAudios"
        elif ext == "so":
            ptype = "preloadedWasm"

        src = f"{ptype}[{repr(src)}]"
        dst = f"{ptype}[{repr(dst)}]"
        swap = f"({src}={dst}) && delete {dst}"
        embed.js(swap, -1)
        # print(swap)


def run_main(PyConfig, loaderhome=None, loadermain="main.py"):
    global ROOTDIR
    global preloadedWasm, preloadedImages, preloadedAudios

    if loaderhome:
        pdb(f"241: appdir mapped to {loaderhome} by loader")
        ROOTDIR = str(loaderhome)
    #
    #    # simulator won't run javascript for now
    #    if not hasattr(embed, "run_script"):
    #        pdb("246: no js engine")
    #        return False
    #
    #    # do not do stuff if not called properly from our js loader.
    #    if PyConfig.executable is None:
    #        # running in sim
    #        pdb("252: running in simulator")
    #        return False
    #
    #    #sys.executable = PyConfig.executable or "python"

    preloadedWasm = "so"
    preloadedImages = "png jpeg jpg gif"
    preloadedAudios = "wav ogg mp4"

    def preload_apk(p=None):
        global preloading, prelist, ROOTDIR
        global explore, preloadedWasm, preloadedImages, preloadedAudios
        ROOTDIR = p or ROOTDIR
        if os.path.isdir(ROOTDIR):
            os.chdir(ROOTDIR)
        else:
            pdb(f"cannot chdir to {ROOTDIR=}")
            return

        ROOTDIR = os.getcwd()
        LSRC = len(ROOTDIR) + 1
        preloading = -1
        prelist = {}

        sys.path.insert(0, ROOTDIR)

        explore(ROOTDIR)

        if preloading < 0:
            pdb(f"{ROOTDIR=}")
            pdb(f"{os.getcwd()=}")

        print(f"284: assets found :", preloading)
        if not preloading:
            embed.run()

        return True

    import aio

    if PyConfig.interactive:
        import aio.clock

        aio.clock.start(x=80)

        # org.python REPL no preload !
        preload = sys.argv[0] != "org.python"
    else:
        # org.python script
        preload = True

    pdb(f"274: preload status {preload}")

    if preload and preload_apk():

        def fix_preload_table_apk():
            global fix_preload_table, ROOTDIR
            fix_preload_table()
            os.chdir(ROOTDIR)
            sys.path.insert(0, ROOTDIR)
            if loadermain:
                if os.path.isfile("main.py"):
                    print(f"315: running {ROOTDIR}/{loadermain} for {sys.argv[0]} (deferred)")
                    aio.defer(execfile, [f"{ROOTDIR}/{loadermain}"], {})
                else:
                    pdb(f"no {loadermain} found for {sys.argv[0]} in {ROOTDIR}")
                # aio.defer(embed.prompt, (), {}, delay=2000)

    # C should unlock aio loop when preload count reach 0.

    else:

        def fix_preload_table_apk():
            global fix_preload_table_apk, ROOTDIR
            pdb("no assets preloaded")
            os.chdir(ROOTDIR)
            # aio.defer(embed.prompt, (), {})

        # unlock embed looper because no preloading
        embed.run()

    aio.defer(fix_preload_table_apk, (), {}, delay=1000)

    if not aio.started:
        aio.started = True
        aio.create_task(EventTarget.process())
    else:
        print("341: EventTarget delayed by loader")


# ===============================================================================================
# platform


def rcp(source, destination):
    import urllib
    import urllib.request

    filename = Path(destination)
    filename.parent.mkdir(parents=True, exist_ok=True)
    urllib.request.urlretrieve(source_url, str(filename))



import sys

DEBUG = False

import aio.prepro

aio.prepro.DEBUG = DEBUG

#
platform_impl = False

# ================== leverage known python implementations ====================

# always come down to upy choices because cpython syntax can more easily be adapted


if not defined("__UPY__"):
    define("__UPY__", hasattr(sys.implementation, "_mpy"))

    if not __UPY__:
        # setup exception display with same syntax as upy
        import traceback

        def print_exception(e, out=sys.stderr, **kw):
            kw["file"] = out
            traceback.print_exc(**kw)

        sys.print_exception = print_exception
        del print_exception


if not defined("__WASM__"):
    try:
        # that sym cannot be overloaded in the simulator
        if __import__("os").uname().machine.startswith("wasm"):
            import __WASM__
        else:
            __WASM__ = False
    except AttributeError:
        # upy does not have os.uname
        __WASM__ = True

    define("__WASM__", __WASM__)


if not defined("__wasi__"):
    if sys.platform == "wasi":
        import __wasi__
    else:
        __wasi__ = False

define("__wasi__", __wasi__)


# this *is* the cpython way
if hasattr(sys, "getandroidapilevel"):
    platform_impl = defined("__ANDROID__")
    if not platform_impl:
        print("importing platform_impl __ANDROID__")
        try:
            import __ANDROID__ as platform_impl
        except Exception as e:
            if hasattr(sys, "print_exception"):
                sys.print_exception(e)
            else:
                __import__("traceback").print_exc()

            pdb("__ANDROID__ failed to load, assuming simulator instead of error :")
            del platform_impl

            # fake it
            platform_impl == __import__("__main__")
        define("__ANDROID__", platform_impl)
    try:
        __ANDROID_API__
    except:
        defined("__ANDROID_API__", sys.getandroidapilevel())

define("__ANDROID__", platform_impl)


if sys.platform == "emscripten":
    platform_impl = defined("__EMSCRIPTEN__")
    if not platform_impl:
        try:
            import __EMSCRIPTEN__ as platform_impl
        except Exception as e:
            if hasattr(sys, "print_exception"):
                sys.print_exception(e)
            else:
                __import__("traceback").print_exc()

            pdb("__EMSCRIPTEN__ failed to load, assuming simulator instead of error :")
            del platform_impl

            # fake it
            platform_impl == __import__("__main__")

define("__EMSCRIPTEN__", platform_impl)


driver = defined("embed")
try:
    if not driver:
        import embed as driver
except:
    # else use the simulator defined platform_impl value as the embed.
    driver = platform_impl
if driver:
    # just in case it was not a module
    sys.modules.setdefault("embed", driver)

    try:
        # check it the embedding module was finished for that platform_impl.
        # the least shoulbe syslog ( js console / adb logcat )
        driver.log
    except:
        pdb(
            """\
    WARNING: embed softrt/syslog functions not found
    WARNING: also not in __main__ or simulator provided platform_impl module
    """
        )
        driver.enable_irq = print
        driver.disable_irq = print
        driver.log = print

    define("embed", driver)

    platform_impl.init_platform(driver)
    sys.modules["platform"] = platform_impl


if not __UPY__:
    # setup exception display with same syntax as upy
    import traceback

    def print_exception(e, out=sys.stderr, **kw):
        kw["file"] = out
        traceback.print_exc(**kw)

    sys.print_exception = print_exception
    del print_exception


scheduler = None
simulator = None



#!pythonrc.py

import os, sys, json, builtins


# to be able to access aio.cross.simulator
import aio
import aio.cross

# placeholder until v1.0
sys.modules["pygbag"] = aio

import time
import inspect
from pathlib import Path
import json


PYCONFIG_PKG_INDEXES_DEV = ["http://localhost:<port>/archives/repo/"]
PYCONFIG_PKG_INDEXES = ["https://pygame-web.github.io/archives/repo/"]

# the sim does not preload assets and cannot access currentline
# unless using https://github.com/pmp-p/aioprompt/blob/master/aioprompt/__init__.py
# or a thread

if not defined("undefined"):

    class sentinel:
        def __bool__(self):
            return False

        def __repr__(self):
            return "∅"

        def __nonzero__(self):
            return 0

        def __call__(self, *argv, **kw):
            if len(argv) and argv[0] is self:
                return True
            print("Null Pointer Exception")

    define("undefined", sentinel())
    del sentinel

    define("false", False)
    define("true", True)

    # fix const without writing const in that .py because of faulty micropython parser.
    exec("__import__('builtins').const = lambda x:x", globals(), globals())


def overloaded(i, *attrs):
    for attr in attrs:
        if attr in vars(i.__class__):
            if attr in vars(i):
                return True
    return False


builtins.overloaded = overloaded


def DBG(*argv):
    if PyConfig.dev_mode > 0:
        print(*argv)


try:
    # mpy already has execfile
    execfile
except:

    def execfile(filename):
        imports = []

        # this buggy parser is for implementations that do not have ast module.
        # and should not be used with cpython
        with __import__("tokenize").open(str(filename)) as f:
            __prepro = []
            myglobs = ["setup", "loop", "main"]
            tmpl = []

            for l in f.readlines():
                testline = l.split("#")[0].strip(" \r\n,\t")

                if testline.startswith("global ") and (
                    testline.endswith(" setup") or testline.endswith(" loop") or testline.endswith(" main")
                ):
                    tmpl.append([len(__prepro), l.find("g")])
                    __prepro.append("#globals")
                    continue

                elif testline.startswith("import "):
                    testline = testline.replace("import ", "").strip()
                    for elem in map(str.strip, testline.split(",")):
                        elem = elem.split(" as ")[0]
                        if not elem in imports:
                            imports.append(elem)

                elif testline.startswith("from "):
                    testline = testline.replace("from ", "").strip()
                    elem = testline.split(" import ")[0].strip()
                    if not elem in imports:
                        imports.append(elem)

                __prepro.append(l)

                if l[0] in ("""\n\r\t'" """):
                    continue

                if not l.find("=") > 0:
                    continue

                l = l.strip()

                if l.startswith("def "):
                    continue
                if l.startswith("class "):
                    continue

                # maybe found a global assign
                varname = l.split("=", 1)[0].strip(" []()")

                for varname in map(str.strip, varname.split(",")):
                    if varname.find(" ") > 0:
                        continue

                    # it's a comment on an assign !
                    if varname.find("#") >= 0:
                        continue

                    # skip attr assign
                    if varname.find(".") > 0:
                        continue

                    # not a tuple assign
                    if varname.find("(") > 0:
                        continue

                    # not a list assign
                    if varname.find("[") > 0:
                        continue

                    # TODO handle (a,)=(0,) case types

                    if not varname in myglobs:
                        myglobs.append(varname)

            myglob = f"global {', '.join(myglobs)}\n"

            # for helping fixing freshly ported code
            if aio.cross.simulator:
                print(myglob)

            for mark, indent in tmpl:
                __prepro[mark] = " " * indent + myglob

            def dump_code():
                nonlocal __prepro
                print()
                print("_" * 70)
                for i, l in enumerate(__prepro):
                    print(str(i).zfill(5), l, end="")
                print("_" * 70)
                print()

            # if aio.cross.simulator:
            #    dump_code()

            # use of globals() is only valid in __main__ scope
            # we really want the module __main__ dict here
            # whereever from we are called.
            __main__ = __import__("__main__")
            __main__dict = vars(__main__)
            __main__dict["__file__"] = str(filename)
            try:
                code = compile("".join(__prepro), str(filename), "exec")
            except SyntaxError as e:
                # if not aio.cross.simulator:
                dump_code()
                sys.print_exception(e)
                code = None

            if code:
                print(f"180: imports: {imports}")
                exec(code, __main__dict, __main__dict)

        return __import__("__main__")

    define("execfile", execfile)


try:
    PyConfig
except NameError:
    PyConfig = None

# in simulator there's no PyConfig
# would need to get one from live cpython
if PyConfig is None:
    # TODO: build a pyconfig extracted from C here
    PyConfig = {}
    PyConfig["dev_mode"] = 1
    PyConfig["run_filename"] = "main.py"

    # TODO: use location of python js module.
    if __UPY__:
        PyConfig["executable"] = "upy"
    else:
        PyConfig["executable"] = sys.executable

    PyConfig["interactive"] = 1
    print(" - running in wasm simulator - ")
    aio.cross.simulator = True
else:
    PyConfig["pkg_repolist"] = []

    aio.cross.simulator = False
    sys.argv.clear()
    sys.argv.extend(PyConfig.pop("argv", []))

    sys.executable = PyConfig["executable"]
    sys.orig_argv.clear()

    sys.orig_argv.append(sys.executable)

    # env is passed in orig_argv ?ENV1=V1&ENV2=V2#
    for arg in PyConfig["orig_argv"]:
        if "=" not in arg:
            sys.orig_argv.append(arg)
        else:
            k, v = arg.split("=", 1)
            os.environ[k] = v

    home = f"/home/{os.environ.get('USER','web_user')}"
    if home != "/home/web_user":
        # in case user name is not fs compatible
        try:
            os.rename("/home/web_user", home)
        except:
            home = "/home/web_user"

    os.environ["HOME"] = home
    os.environ["APPDATA"] = home
    del home


PyConfig["imports_ready"] = False
PyConfig["pygbag"] = 0

PyConfig.setdefault("user_site_directory", 0)


class shell:
    # pending async tasks
    coro = []

    # async top level instance compiler/runner
    runner = None
    is_interactive = None

    if aio.cross.simulator or not len(sys.argv):
        ROOT = os.getcwd()
        HOME = os.getcwd()
    else:
        ROOT = f"/data/data/{sys.argv[0]}"
        HOME = f"/data/data/{sys.argv[0]}/assets"

    pgzrunning = None

    @classmethod
    def mktemp(cls, suffix=""):
        return aio.filelike.mktemp(suffix)

    @classmethod
    def cat(cls, *argv):
        """dump binary file content"""
        for fn in map(str, argv):
            with open(fn, "rb") as out:
                print(out.read())

    @classmethod
    def more(cls, *argv):
        """dump text file content"""
        for fn in map(str, argv):
            with open(fn, "r") as out:
                print(out.read())

    @classmethod
    def pp(cls, *argv):
        """pretty print objects via json"""
        for obj in argv:
            obj = eval(obj, vars(__import__("__main__")))
            if isinstance(obj, platform.Object_type):
                obj = json.loads(platform.window.JSON.stringify(obj))
            yield json.dumps(obj, sort_keys=True, indent=4)

    @classmethod
    def ls(cls, *argv):
        """list directory content"""
        if not len(argv):
            argv = ["."]
        for arg in map(str, argv):
            for out in sorted(os.listdir(arg)):
                print(out)

    @classmethod
    def reset(cls, *argv, **kw):
        ESC("c")

    @classmethod
    def pg_init(cls):
        import pygame

        if pygame.display.get_init():
            return pygame.display.get_surface()
        screen = pygame.display.set_mode([cls.screen_width, cls.screen_height])
        return screen

    @classmethod
    def find(cls, *argv):
        from pathlib import Path

        if not len(argv):
            argv = [os.getcwd()]
        for root in argv:
            root = Path(root)
            for current, dirnames, filenames in os.walk(root):
                dirname = root.joinpath(Path(current))
                for file in filenames:
                    yield str(dirname / file)

    @classmethod
    def grep(cls, match, *argv):
        for arg in argv:
            if arg.find(match) > 0:
                yield arg

    @classmethod
    def clear(cls, *argv, **kw):
        """clear terminal screen"""
        import pygame

        screen = cls.pg_init()
        screen.fill((0, 0, 0))
        pygame.display.update()

    @classmethod
    def display(cls, *argv, **kw):
        """show images, or last repl pygame surface from _"""
        import pygame

        if not len(argv):
            surf = _
        else:
            arg = argv[-1]
            ext = arg.lower()
            if ext.endswith(".b64"):
                import base64

                ext = arg[:-4]
                with open(arg, "rb") as infile:
                    arg = arg[:-4]
                    with open(arg, "wb") as outfile:
                        base64.decode(infile, outfile)

            if ext.endswith(".six"):
                cls.more(arg)
                return

            if ext.endswith(".bmp"):
                surf = pygame.image.load_basic(arg)
            else:
                surf = pygame.image.load(arg)

        screen = cls.pg_init()
        screen.blit(surf, (1, 1))
        pygame.display.update()

    @classmethod
    def mkdir(cls, *argv):
        exist_ok = "-p" in argv
        for arg in map(str, argv):
            if arg == "-p":
                continue
            os.makedirs(arg, exist_ok=exist_ok)

    @classmethod
    def rx(cls, *argv, **env):
        for arg in map(str, argv):
            if arg.startswith("-"):
                continue
            platform.window.MM.download(arg)
            yield f"file {arg} sent"
        return True

    @classmethod
    async def async_pgzrun(cls, *argv, **env):
        await __import__("pgzero").runner.PGZeroGame(__import__("__main__")).async_run()

    @classmethod
    def pgzrun(cls, *argv, **env):
        import pgzero
        import pgzero.runner

        pgzt = pgzero.runner.PGZeroGame(__import__("__main__")).async_run()
        asyncio.create_task(pgzt)
        return True

    @classmethod
    def wget(cls, *argv, **env):
        import urllib.request

        filename = None
        for arg in map(str, argv):
            if arg.startswith("-O"):
                filename = arg[2:].strip()
                yield f'saving to "{filename}"'
                break

        for arg in map(str, argv):
            if arg.startswith("-O"):
                continue
            fn = filename or str(argv[0]).rsplit("/")[-1]
            try:
                filename, _ = urllib.request.urlretrieve(str(arg), filename=fn)
            except Exception as e:
                yield e

        return True

    @classmethod
    def env(cls, *argv):
        for k in os.environ:
            yield f"{k}={os.environ[k]}"
        return True

    @classmethod
    def pwd(cls, *argv):
        print(os.getcwd())

    # only work if pkg name == dist name
    @classmethod
    async def pip(cls, *argv):
        for arg in argv:
            if arg == "install":
                continue
            import aio.pep0723

            # yield f"attempting to install {arg}"
            await aio.pep0723.pip_install(arg)

    @classmethod
    def cd(cls, *argv):
        if len(argv):
            os.chdir(argv[-1])
        else:
            os.chdir(cls.HOME)
        print("[ ", os.getcwd(), " ]")

    @classmethod
    def sha256sum(cls, *argv):
        import hashlib

        for arg in map(str, argv):
            sha256_hash = hashlib.sha256()
            with open(arg, "rb") as f:
                # Read and update hash string value in blocks of 4K
                for byte_block in iter(lambda: f.read(4096), b""):
                    sha256_hash.update(byte_block)
                hx = sha256_hash.hexdigest()
                yield f"{hx}  {arg}"

    @classmethod
    def spawn(cls, cmd, *argv, **env):
        # TODO extract env from __main__ snapshot
        if cmd.endswith(".py"):
            if cls.pgzrunning:
                print("a program is already running, using 'stop' cmd before retrying")
                cls.stop()
                cls.pgzrunning = None
                args = [cmd]
                args.extend(argv)
                aio.defer(cls.spawn, args, env, delay=500)
            else:
                execfile(cmd)
            return True
        return False

    @classmethod
    def umask(cls, *argv, **kw):
        yield oct(os.umask(0))
        return True

    @classmethod
    def chmod(cls, *argv, **kw):
        def _current_umask() -> int:
            mask = os.umask(0)
            os.umask(mask)
            return mask

        for arg in argv:
            if arg.startswith("-"):
                continue
            mode = 0o777 & ~_current_umask() | 0o111
            print(f"{mode=}")
            os.chmod(arg, mode)

    @classmethod
    def unzip(cls, *argv, **env):
        import zipfile

        for zip in argv:
            with zipfile.ZipFile(zip, "r") as zip_ref:
                zip_ref.printdir()
                zip_ref.extractall(os.getcwd())

    @classmethod
    def install(cls, *argv, **env):
        import aio.pep0723

        for pkg_file in argv:
            try:
                aio.pep0723.install(pkg_file)
                yield f"{pkg_file} installed"
            except (IOError, zipfile.BadZipFile):
                pdb("397: invalid package", pkg_file)
            except Exception as ex:
                sys.print_exception(ex)

    @classmethod
    def dll(cls, *argv):
        cdll = __import__("ctypes").CDLL(None)
        sym = getattr(cdll, argv[0])
        print("symbol :", sym)
        print(sym(*argv[1:]))
        return True

    @classmethod
    def strace(cls, *argv, **env):
        import aio.tracer

        print("497: trace on")
        sys.settrace(aio.tracer.calls)
        return True

    @classmethod
    def mute(cls, *argv, **env):
        try:
            pygame.mixer.music.unload()
            yield "music muted"
        except:
            pass

    @classmethod
    def debug(cls, *argv, **env):
        try:
            platform.window.debug()
            yield f"debug mode : on, canvas divider {window.python.config.gui_debug}"
        except:
            pass

    @classmethod
    def help(cls, *objs):
        print(
            """
pygbag shell help
________________________
"""
        )
        if not len(objs):
            objs = [cls]
        for obj in objs:
            for cmd, item in vars(obj).items():
                if isinstance(item, str):
                    continue
                if cmd[0] != "_" and item.__doc__:
                    print(cmd, ":", item.__doc__)
                    print()

    # TODO: use run interactive c-api to run this one.
    @classmethod
    def run(cls, *argv, **env):
        __main__ = __import__("__main__")
        __main__dict = vars(__main__)

        builtins._ = undefined
        cmd = " ".join(argv)

        try:
            time_start = time.time()
            code = compile("builtins._ =" + cmd, "<stdin>", "exec")
            exec(code, __main__dict, __main__dict)
            if builtins._ is undefined:
                return True
            if aio.iscoroutine(_):

                async def run(coro):
                    print(f"async[{cmd}] :", await coro)
                    print(f"time[{cmd}] : {time.time() - time_start:.6f}")

                aio.create_task(run(_), name=cmd)
            else:
                print(builtins._)
                print(f"time[{cmd}] : {time.time() - time_start:.6f}")
                return True
        except SyntaxError as e:
            # try run a file or cmd
            return cls.parse_sync(argv, env)
        return False

    time = run

    @classmethod
    def ps(cls, *argv, **env):
        for t in aio.all_tasks():
            print(t)
        return True

    @classmethod
    def stop(cls, *argv, **env):
        aio.exit = True
        # pgzrun will reset to None next exec
        if not cls.pgzrunning:
            # pgzrun does its own cleanup call
            aio.defer(aio.recycle.cleanup, (), {}, delay=500)
            aio.defer(platform.prompt, (), {}, delay=800)

    @classmethod
    def uptime(cls, *argv, **env):
        import asyncio, platform

        if platform.is_browser:

            def load_display(ft):
                avg = sum(ft) / len(ft)
                try:
                    platform.window.load_avg.innerText = "{:.4f}".format(avg)
                    platform.window.load_min.innerText = "{:.4f}".format(min(ft))
                    platform.window.load_max.innerText = "{:.4f}".format(max(ft))
                    return True
                except:
                    pdb("366:uptime: window.load_* widgets not found")
                    return False

            async def perf_index():
                ft = [0.00001] * 60 * 10
                while not aio.exit:
                    ft.pop(0)
                    ft.append(aio.spent / 0.016666666666666666)
                    if not (aio.ticks % 60):
                        if not load_display(ft):
                            break
                    await asyncio.sleep(0)

            aio.create_task(perf_index())
        else:
            print(f"last frame : {aio.spent / 0.016666666666666666:.4f}")

    @classmethod
    async def preload_code(cls, code, callback=None, hint=""):
        # get a relevant list of modules likely to be imported
        PyConfig.dev_mode = 1
        DBG(f"632: preload_code({len(code)=} {hint=}")

        import aio
        import aio.pep0723
        from aio.pep0723 import Config

        if not aio.cross.simulator:
            # env path is set by pep0723
            sconf = __import__("sysconfig").get_paths()
            env = Path(sconf["purelib"])

            if not len(Config.repos):
                if not len(Config.PKG_INDEXES):
                    Config.PKG_INDEXES = PyConfig.pkg_indexes
                for cdn in Config.PKG_INDEXES:
                    async with platform.fopen(Path(cdn) / Config.REPO_DATA) as source:
                        Config.repos.append(json.loads(source.read()))

                DBG("650: FIXME (this is pyodide maintened stuff, use (auto)PEP723 asap)")
                print("651: referenced packages :", len(Config.repos[0]["packages"]))

            DBG(f"654: aio.pep0723.check_list {aio.pep0723.env=}")
            deps = await aio.pep0723.check_list(code)

            DBG(f"656: aio.pep0723.pip_install {deps=}")

            # auto import plumbing to avoid rely too much on import error
            maybe_wanted = list(TopLevel_async_handler.list_imports(code, file=None, hint=hint))
            DBG(f"635: {maybe_wanted=}")
            for dep in maybe_wanted:
                if not dep in deps:
                    deps.append(dep)

            for dep in deps:
                await aio.pep0723.pip_install(dep)

            aio.pep0723.do_patches()

        PyConfig.imports_ready = True
        return True

    @classmethod
    def interactive(cls, prompt=False):
        if prompt:
            aio.toplevel.handler.mute_state = False
            aio.toplevel.handler.muted = False

        if cls.is_interactive:
            return
        # if you don't reach that step
        # your main.py has an infinite sync loop somewhere !
        DBG("651: starting EventTarget in a few seconds")

        print()
        aio.toplevel.handler.instance.banner()

        aio.create_task(platform.EventTarget.process())
        cls.is_interactive = True

        if not shell.pgzrunning:
            # __main__@stdin has no __file__
            if hasattr(__import__("__main__"), "__file__"):
                del __import__("__main__").__file__
            if prompt:
                cls.runner.prompt()
        else:
            shell.pgzrun()

    @classmethod
    async def runpy(cls, main, *args, **kw):
        def check_code(file_name):
            nonlocal code
            maybe_sync = False
            has_pygame = False
            with open(file_name, "r") as code_file:
                code = code_file.read()
                code = code.rsplit(aio.toplevel.handler.HTML_MARK, 1)[0]

                # do not check site/final/packed code
                # preload code must be fully async and no pgzero based
                if aio.toplevel.handler.muted:
                    return True

                if code[0:320].find("#!pgzrun") >= 0:
                    shell.pgzrunning = True

                if code.find("asyncio.run") < 0:
                    DBG("606: possibly synchronous code found")
                    maybe_sync = True

                has_pygame = code.find("display.flip(") > 0 or code.find("display.update(") > 0

                if maybe_sync and has_pygame:
                    DBG("694: possibly synchronous+pygame code found")
                    return False
            return True

        code = ""
        shell.pgzrunning = None

        DBG(f"690: : runpy({main=})")
        # REMOVE THAT IT SHOULD BE DONE IN SIM ANALYSER AND HANDLED PROPERLY
        if not check_code(main):
            for base in ("pygame", "pg"):
                for func in ("flip", "update"):
                    block = f"{base}.display.{func}()"
                    code = code.replace(block, f"{block};await asyncio.sleep(0)")

        # fix cwd to match a run of main.py from its folder
        realpath = str(main)
        if realpath[0] not in "./":
            realpath = str(Path.cwd() / main)
        __import__("__main__").__file__ = str(realpath)
        cls.HOME = Path(realpath).parent
        os.chdir(cls.HOME)

        # TODO: should be $0 / sys.argv[0] from there and while running
        kw.setdefault("hint", main)
        # get requirements
        await cls.preload_code(code, **kw)

        # get an async executor to catch import errors
        if aio.toplevel.handler.instance:
            DBG("715: starting shell")
            aio.toplevel.handler.instance.start_console(shell)
        else:
            pdb("718: no async handler loader, starting a default async console")
            shell.debug()
            await aio.toplevel.handler.start_toplevel(platform.shell, console=True)

        # TODO: check if that thing really works
        if shell.pgzrunning:
            DBG("728 : pygame zero detected")
            __main__ = __import__("__main__")
            sys._pgzrun = True
            sys.modules["pgzrun"] = type(__main__)("pgzrun")
            import pgzrun

            pgzrun.go = lambda: None
            cb = kw.pop("callback", None)
            await aio.toplevel.handler.async_imports(cb, "pygame.base", "pgzero", "pyfxr", **kw)
            import pgzero
            import pgzero.runner

            pgzero.runner.prepare_mod(__main__)

        # finally eval async
        aio.toplevel.handler.instance.eval(code)

        # go back to prompt
        if not aio.toplevel.handler.muted:
            print("going interactive")
            DBG("746: TODO detect input/print to select repl debug")
            cls.interactive()

        return code

    @classmethod
    async def source(cls, main, *args, **kw):
        # this is not interactive turn off prompting
        aio.toplevel.handler.muted = True
        try:
            return await cls.runpy(main, *args, **kw)
        finally:
            aio.toplevel.handler.muted = aio.toplevel.handler.mute_state

    @classmethod
    def parse_sync(shell, line, **env):
        catch = True
        for cmd in line.strip().split(";"):
            cmd = cmd.strip()
            if cmd.find(" ") > 0:
                cmd, args = cmd.split(" ", 1)
                args = args.split(" ")
            else:
                args = ()

            if hasattr(shell, cmd):
                fn = getattr(shell, cmd)

                try:
                    if inspect.isgeneratorfunction(fn):
                        for _ in fn(*args):
                            print(_)
                    elif inspect.iscoroutinefunction(fn):
                        aio.create_task(fn(*args))
                    elif inspect.isasyncgenfunction(fn):
                        print("asyncgen N/I")
                    elif inspect.isawaitable(fn):
                        print("awaitable N/I")
                    else:
                        fn(*args)

                except Exception as cmderror:
                    print(cmderror, file=sys.stderr)
            elif cmd.endswith(".py"):
                shell.coro.append(shell.source(cmd, *args, **env))
            else:
                catch = undefined
        return catch

    @classmethod
    async def exec(cls, sub, **env):
        if inspect.isgenerator(sub):
            for _ in sub:
                print(_)
            return
        elif inspect.isgeneratorfunction(sub):
            for _ in sub(**env):
                print(_)
            return
        elif inspect.iscoroutinefunction(sub):
            await sub(*args)
            return

        from collections.abc import Iterator

        if isinstance(sub, Iterator):
            for _ in sub:
                print(_)
            return
        elif isinstance(
            sub,
            (
                str,
                Path,
            ),
        ):
            # subprocess
            return cls.parse_sync(sub, **env)
        else:
            await sub


import os

os.shell = shell
builtins.shell = shell
# end shell


if __UPY__:
    import types

    class SimpleNamespace:
        def __init__(self, **kwargs):
            for k, v in kwargs.items():
                setattr(self, k, v)

        def __repr__(self):
            keys = sorted(self.__dict__)
            items = ("{}={!r}".format(k, self.__dict__[k]) for k in keys)
            return "{}({})".format(type(self).__name__, ", ".join(items))

        def __eq__(self, other):
            return self.__dict__ == other.__dict__

    types.SimpleNamespace = SimpleNamespace
else:
    from types import SimpleNamespace

import builtins

builtins.PyConfig = SimpleNamespace(**PyConfig)
del PyConfig


# make simulations same each time, easier to debug
import random

random.seed(1)

import __EMSCRIPTEN__ as platform

platform.shell = shell
import aio.filelike

platform.fopen = aio.filelike.fopen
platform.sopen = aio.filelike.sopen


if not aio.cross.simulator:

    def fix_url(maybe_url):
        url = str(maybe_url)
        if url.startswith("http://"):
            pass
        elif url.startswith("https://"):
            pass
        elif url.startswith("https:/"):
            url = "https:/" + url[6:]
        elif url.startswith("http:/"):
            url = "http:/" + url[5:]
        return url

    platform.fix_url = fix_url

    del fix_url

    def apply_patches():
        # use shell generators instead of subprocesses
        # ==========================================================

        import os

        def popen(iterator, **kw):
            import io

            kw.setdefault("file", io.StringIO(newline="\r\n"))
            for line in iterator:
                print(line, **kw)
            kw["file"].seek(0)
            return kw["file"]

        os.popen = popen

        # add real browser functions
        # ===========================================================

        import webbrowser

        def browser_open(url, new=0, autoraise=True):
            platform.window.open(url, "_blank")

        def browser_open_new(url):
            return browser_open(url, 1)

        def browser_open_new_tab(url):
            return browser_open(url, 2)

        webbrowser.open = browser_open
        webbrowser.open_new = browser_open_new
        webbrowser.open_new_tab = browser_open_new_tab

        # extensions

        def browser_open_file(target=None, accept="*"):
            if target:
                platform.EventTarget.addEventListener(window, "upload", target)
            platform.window.dlg_multifile.click()

        webbrowser.open_file = browser_open_file

        # merge emscripten browser module here ?
        # https://rdb.name/panda3d-webgl.md.html#supplementalmodules/asynchronousloading
        #

        # use bad and deprecated sync XHR for urllib
        # ============================================================

        import urllib
        import urllib.request

        def urlretrieve(maybe_url, filename=None, reporthook=None, data=None):
            url = __EMSCRIPTEN__.fix_url(maybe_url)
            filename = str(filename or f"/tmp/uru-{aio.ticks}")
            rc = platform.window.python.DEPRECATED_wget_sync(str(url), filename)
            if rc == 200:
                return filename, []
            raise Exception(f"urlib.error {rc}")

        urllib.request.urlretrieve = urlretrieve

    if (__WASM__ and __EMSCRIPTEN__) or platform.is_browser:
        port = "443"

        # pygbag mode
        if platform.window.location.href.find("//localhost:") > 0:
            port = str(platform.window.location.port)

            # pygbag developer mode ( --dev )
            if ("-i" in PyConfig.orig_argv) or (port == "8666"):
                PyConfig.dev_mode = 1
                print(sys._emscripten_info)

            PyConfig.pygbag = 1
        else:
            PyConfig.pygbag = 0

        if (PyConfig.dev_mode > 0) or PyConfig.pygbag:
            # in pygbag dev mode use local repo
            PyConfig.pkg_indexes = []
            for idx in PYCONFIG_PKG_INDEXES_DEV:
                redirect = idx.replace("<port>", port)
                PyConfig.pkg_indexes.append(redirect)

            print("807: DEV MODE ON", PyConfig.pkg_indexes)
        else:
            # address cdn
            PyConfig.pkg_indexes = PYCONFIG_PKG_INDEXES

        from platform import window, document, ffi

        apply_patches()

    del apply_patches

    # convert a emscripten FS path to a blob url
    # TODO: weakmap and GC collect
    def File(path):
        return platform.window.blob(str(path))

    # =================== async import , async console ===================================

    import os

    # set correct umask ( emscripten default is 0 )
    if hasattr(os, "umask"):
        os.umask(0o022)  # already done in aio.toplevel
        import zipfile
    else:
        pdb("1010: missing os.umask")
        pdb("1011: missing zipfile")

    import aio.toplevel

    # import ast
    from pathlib import Path

    class TopLevel_async_handler(aio.toplevel.AsyncInteractiveConsole):
        # be re entrant
        import_lock = []

        mute_state = False

        HTML_MARK = '"' * 3 + " # BEGIN -->"

        repos = []

        may_need = []
        ignore = ["ctypes", "distutils", "installer", "sysconfig"]
        ignore += ["python-dateutil", "matplotlib-pyodide"]
        # ???
        ignore += ["pillow", "fonttools"]

        # for ursina
        # ignore +=  ["ursina","gltf","pyperclip","screeninfo"]

        manual_deps = {
            "matplotlib": ["numpy", "six", "cycler", "PIL", "pygame-ce"],
            "bokeh": ["numpy", "yaml", "typing_extensions", "jinja2", "markupsafe"],
            "igraph": ["texttable"],
            "pygame_gui": ["pygame.base", "i18n"],
            "ursina": ["numpy", "screeninfo", "gltf", "PIL", "pyperclip", "panda3d"],
        }

        missing_fence = []

        from pathlib import Path

        # repodata = "repodata.json"

        def eval(self, source):
            for count, line in enumerate(source.split("\n")):
                if not count:
                    if line.startswith("<"):
                        self.buffer.append(f"#{line}")
                        continue
                self.buffer.append(line)

            if count:
                self.line = None
                self.buffer.insert(0, "#")
            # self.buffer.append("")
            DBG(f"1039: {count} lines queued for async eval")

        @classmethod
        def scan_imports(cls, code, filename, load_try=False, hint=""):
            import aio.pep0723
            import ast

            required = []
            try:
                root = ast.parse(code, filename)
            except SyntaxError as e:
                print("_" * 40)
                print("1111:", filename, hint)
                print("_" * 40)
                for count, line in enumerate(code.split("\n")):
                    print(str(count).zfill(3), line)
                sys.print_exception(e)
                return required

            for node in ast.walk(root):
                if isinstance(node, ast.Import):
                    module = []
                elif isinstance(node, ast.ImportFrom):
                    module = node.module.split(".")
                else:
                    continue

                for n in node.names:
                    if len(module):
                        mod = module[0] or n.name.split(".")[0]
                    else:
                        mod = n.name.split(".")[0]

                    mod = aio.pep0723.Config.mapping.get(mod, mod)

                    if mod in cls.ignore:
                        continue

                    if mod in cls.may_need:
                        continue

                    if mod in sys.modules:
                        continue

                    if load_try:
                        try:
                            __import__(mod)
                            continue
                        except (ModuleNotFoundError, ImportError):
                            pass

                    if not mod in required:
                        required.append(mod)

            DBG(f"1153: scan_imports {hint=} {filename=} {len(code)=} {required}")
            return required

        @classmethod
        def list_imports(cls, code=None, file=None, hint=""):
            import aio.pep0723

            if not len(aio.pep0723.Config.pkg_repolist):
                print(
                    """
1170: pep0723 REPOSITORY MISSING
"""
                )
            else:
                DBG(
                    f"""
1175: list_imports {len(code)=} {file=} {hint=}")
{aio.pep0723.Config.pkg_repolist[0]['-CDN-']=}

"""
                )

            if code is None:
                if file:
                    with open(file) as fcode:
                        code = fcode.read()
                else:
                    code = ""

            file = file or "<stdin>"

            for want in cls.scan_imports(code, file, hint=hint):
                # DBG(f"1114: requesting module {want=} for {file=} ")
                repo = None
                for repo in aio.pep0723.Config.pkg_repolist:
                    if want in cls.may_need:
                        DBG(f"1118: skip module {want=} reason: already requested")
                        break

                    if want in sys.modules:
                        DBG(f"1122: skip module {want=} reason: sys.modules")
                        break

                    if want in repo:
                        cls.may_need.append(want)
                        # DBG(f"1127: module {want=} requested")
                        yield want
                        break
                else:
                    if repo:
                        DBG(f"1187: {repo['-CDN-']=} does not provide {want=}")
                    else:
                        print("1189: no pkg repository available")

        # TODO: re order repo on failures
        # TODO: try to download from pypi with
        # https://github.com/brettcannon/mousebender/blob/main/mousebender/simple.py
        # https://peps.python.org/pep-0503/
        # https://wiki.python.org/moin/PyPIJSON

        # TODO: gets deps from pygbag
        # https://github.com/thebjorn/pydeps

        @classmethod
        def import_one(cls, mod, lvl=0):
            wants = []

            if mod in sys.modules:
                return []

            if mod in cls.missing_fence:
                return []
            from aio.pep0723 import Config

            for dep in Config.repos[0]["packages"].get(mod, {}).get("depends", []):
                if dep in cls.ignore:
                    continue

                if dep in cls.missing_fence:
                    continue

                cls.missing_fence.append(dep)

                if lvl < 3:
                    for subdep in cls.imports(mod, lvl=lvl + 1):
                        if not subdep in cls.missing_fence:
                            cls.missing_fence.append(subdep)

            if mod in cls.manual_deps:
                deps = list(cls.manual_deps[mod])
                deps.reverse()
                DBG(
                    f"""
1242:
    added {deps=} for {mod=}
1243:
    {cls.missing_fence=}

"""
                )
                for missing in deps:
                    if missing in cls.missing_fence:
                        continue

                    if missing in wants:
                        continue

                    # no need to request
                    if missing in sys.modules:
                        continue

                    # prio missing
                    wants.insert(0, missing)
                    DBG(f"1108: added {missing=} for {mod=}")

            wants.append(mod)
            return wants

        @classmethod
        def imports(cls, *mods, lvl=0):
            wants = []
            unseen = False
            for mod in mods:
                # get potential sub deps
                for dep in cls.import_one(mod, lvl=lvl):
                    if dep in wants:
                        continue

                    if dep in sys.modules:
                        continue

                    if dep in cls.missing_fence:
                        continue

                    if dep in cls.ignore:
                        continue

                    wants.append(dep)

                if not mod in wants:
                    wants.append(mod)

            return wants

    # end TopLevel_async_handler

    aio.toplevel.handler = TopLevel_async_handler


else:
    pdb("TODO: js simulator")


try:
    shell.screen_width = int(platform.window.canvas.width)
    shell.screen_height = int(platform.window.canvas.height)
except:
    shell.screen_width = 1024
    shell.screen_height = 600

# ======================================================
# patching
# import platform_wasm.todo


# ======================================================
# x10 mouse and xterm stuff
# https://github.com/muesli/termenv/pull/104
# https://xtermjs.org/docs/api/vtfeatures/

if not aio.cross.simulator:

    def ESC(*argv):
        for arg in argv:
            sys.__stdout__.write(chr(0x1B))
            sys.__stdout__.write(arg)
        embed.flush()

    def CSI(*argv):
        for arg in argv:
            ESC(f"[{arg}")

    builtins.ESC = ESC
    builtins.CSI = CSI

try:
    console
except:

    class console:
        def log(*argv, **kw):
            import io

            kw["file"] = io.StringIO(newline="\r\n")
            print(*argv, **kw)
            embed.warn(kw["file"].getvalue())


import aio.recycle

# ============================================================
# DO NOT ADD ANYTHING FROM HERE OR APP RECYCLING WILL TRASH IT

#

LOCK = False


async def import_site(__file__, run=True):
    global LOCK
    if LOCK:
        print("1728: import_site IS NOT RE ENTRANT")
        return

    try:
        LOCK = True
        from pathlib import Path

        embed = False
        hint = "main.py"

        is_py = sys.argv[0].endswith(".py")

        # if not imported by simulator then aio is handled externally
        if "pygbag.aio" not in sys.modules:
            import aio

            sys.modules["pygbag.aio"] = aio

        # if running a script be silent for prompt
        TopLevel_async_handler.mute_state = ".py" in "".join(sys.argv)

        # always start async handler or we could not do imports on import errors.
        await TopLevel_async_handler.start_toplevel(platform.shell, console=True)

        # RUNNING GIVEN DISK FILE with no prompt
        # this is usually the import site given by javascript loader or a template loader (pygbag apk mode)
        # or the user script (script mode).

        if Path(__file__).is_file():
            DBG(f"1755: shell.source({__file__=})")
            await shell.source(__file__)

            # allow to set user site customization network, or embedded js to be processed
            await asyncio.sleep(0)

            if PyConfig.user_site_directory:
                DBG(f"1768: {__file__=} done, giving hand to user_site")
                return __file__
            else:
                DBG(f"1764: {__file__=} done : now trying user sources")
        else:
            DBG(f"1767: {__file__=} NOT FOUND : now trying user sources")

        # NOW CHECK OTHER SOURCES

        # where to retrieve
        import tempfile

        tmpdir = Path(tempfile.gettempdir())

        # maybe a script filename or content passed as frozen config.

        source = getattr(PyConfig, "frozen", "")
        if source:
            if Path(source).is_file():
                source_path = getattr(PyConfig, "frozen_path", "")
                handler = getattr(PyConfig, "frozen_handler", "")
                DBG("1786: embed path", source_path, "will embed", source, "handled by", handler)
                local = tmpdir / "embed.py"
                with open(source, "r") as src:
                    with open(local, "w") as file:
                        file.write("import sys, pygame;from aio.fetch import FS\n")
                        file.write(src.read())

                        # default handler is run() when embedding
                        if not handler:
                            file.write(
                                """
    __main__ = vars().get('run')
    async def main():
        global __main__
        if 'aio.fetch' in sys.modules:
            import aio.fetch
            await aio.fetch.preload()
            await asyncio.sleep(0)
        if __main__:
            await __main__()
    asyncio.run(main())
    """
                            )
                        else:
                            async with platform.fopen(handler) as handle:
                                file.write("\n")
                                file.write(handle.read())
                embed = True
            else:
                print(f"1814: invalid embed {source=}")
                return None

            # file has been retrieved stored in local
        else:
            local = None
            # no embed, try sys.argv[0] first, but main.py can only be a hint.
            # of what to run in an archive

            if sys.argv[0] == "main.py" or not is_py:
                source = PyConfig.orig_argv[-1]
                if is_py:
                    hint = sys.argv[0]
            else:
                source = sys.argv[0]

        DBG(f"1830: {local=} {source=} {is_py=} {hint=}")

        if local is None:
            ext = str(source).rsplit(".")[-1].lower()

            if ext == "py":
                local = tmpdir / source.rsplit("/", 1)[-1]
                await shell.exec(shell.wget(f"-O{local}", source))

            # TODO: test tar.bz2 lzma tar.xz
            elif ext in ("zip", "gz", "tar", "apk", "jar"):
                DBG(f"1841: found archive source {source=}")
                # download and unpack into tmpdir
                fname = tmpdir / source.rsplit("/")[-1]

                if ext in ("apk", "jar"):
                    fname = fname + ".zip"

                async with platfom.fopen(source, "rb") as zipdata:
                    with open(fname, "wb") as file:
                        file.write(zipdata.read())
                import shutil

                shutil.unpack_archive(fname, tmpdir)
                os.unlink(fname)

                # locate for an entry point after decompression
                hint = "/" + hint.strip("/")
                for file in shell.find(tmpdir):
                    if file.find(hint) > 0:
                        local = tmpdir / file
                        break
                DBG("1862: import_site: found ", local)
            else:
                # maybe base64 or frozen code in html.
                ...

        DBG(f"1867: {local=} {source=} {is_py=} {hint=}")

        if local and local.is_file():
            pdir = str(local.parent)
            os.chdir(pdir)
            if "-v" in PyConfig.orig_argv:
                print()
                print("_" * 70)
                with open(local, "r") as source:
                    for i, l in enumerate(source.readlines()):
                        print(str(i).zfill(5), l, end="")
                print()
                print("_" * 70)
                print()

            # TODO: check orig_argv for isolation parameters
            if not pdir in sys.path:
                sys.path.insert(0, pdir)

            if run:
                await shell.runpy(local)
            return str(local)
        else:
            # show why and drop to prompt
            print(f"404: embed={source} or {sys.argv=}")
            shell.interactive(prompt=True)
            return None
    finally:
        LOCK = False



import sys

if sys.platform not in ("emscripten", "wasi") or aio.cross.simulator:
    sleep_delay = 0
else:
    sleep_delay = 0.0155


from pygbag_ui import Tui, TTY, clear

import pygbag_ui as ui
import pygbag_ux as ux

# ux.dim(1280 // 2 , 720 // 2)


class console:
    @classmethod
    def log(self, *argv, **kw):
        import io

        sio = io.StringIO()
        kw["file"] = sio
        print(*argv, **kw)
        ui.clog.append(sio.read())

    @classmethod
    def get(self):
        import platform

        try:
            platform.window.pyconsole.hidden = False
            platform.window.document.body.style.background = "#000000"
        except:
            ...

        import os

        _, LINES = os.get_terminal_size()
        CONSOLE = os.get_console_size()

        # split the display
        if sys.platform not in ("emscripten", "wasi") or aio.cross.simulator:
            LINES = LINES - CONSOLE

        TTY.set_raw(1)

        platform.shell.is_interactive = False
        platform.shell.interactive(True)
        aio.toplevel.handler.muted = False

        clear(LINES, CONSOLE)  # , '>C> ')
        return self



# screen pixels (real, hardware)
WIDTH = 1280  # {{cookiecutter.width}}
HEIGHT = 720  # {{cookiecutter.height}}


def dim(w, h):
    global WIDTH, HEIGHT
    WIDTH = int(w)
    HEIGHT = int(h)


# reference/idealized screen pixels
REFX = 1980
REFY = 1080


def u(real, ref, v):
    if abs(v) < 0.9999999:
        result = int((float(real) / 100.0) * (v * 1000))
        if v < 0:
            return real - result
        return result
    return int((real / ref) * v)


def x(*argv):
    global WIDTH, REFX
    acc = 0
    for v in argv:
        acc += u(WIDTH, REFX, v)
    return acc


def y(*argv):
    global HEIGHT, REFY
    acc = 0
    for v in argv:
        acc += u(HEIGHT, REFY, v)
    return acc


def r(*argv):
    x = ux(argv[0])
    y = uy(argv[1])
    ret = [x, y]
    if len(argv) > 2:
        w = ux(argv[2])
        h = uy(argv[3])
        ret.append(w)
        ret.append(h)
    return ret


from pathlib import Path

# ====================== pygame


def pg_load(fn, resize=None, width=0, height=0, alpha=True):
    import pygame
    from pathlib import Path

    if Path(fn).is_file():
        media = pygame.image.load(fn)
    else:
        media = pygame.image.load(Path(__file__).parent / "offline.png")

    if resize:
        tmp = pygame.transform.smoothscale(media, resize)
        media = tmp
    elif width and height:
        media = pygame.transform.smoothscale(
            media,
            (
                width,
                heigh,
            ),
        )

    try:
        if alpha:
            return media.convert_alpha()
        else:
            return media.convert()
    # offscreen case
    except:
        return media


class vpad:
    X = 0
    Z = 1
    Y = 2
    evx = []
    evy = []
    evz = []
    axis = [evx, evz, evy]

    LZ = 0.5

    @classmethod
    def get_axis(self, n):
        if len(self.axis[n]):
            return self.axis[n].pop(0)
        return 0.0

    @classmethod
    def emit(self, axis, value):
        import pygame

        self.axis[axis].append(float(value))
        ev = pygame.event.Event(pygame.JOYAXISMOTION)
        pygame.event.post(ev)
        return False



import sys
import time
import asyncio
import os

WTF = 1


class tui:
    # use direct access, it is absolute addressing on raw terminal.
    if 0:
        out = sys.__stdout__.write
    else:

        def out(self, *argv, **kw):
            ...

    try:
        LINES = int(os.environ.get("LINES", 200))
    except:
        LINES = 200

    # save cursor
    def __enter__(self):
        self.out("\x1b7", "\x1b[?25l")
        return self

    # restore cursor
    def __exit__(self, *tb):
        self.out("\x1b8", "\x1b[?25h")

    # TODO: limit buffer to LINES
    def __call__(self, *a, **kw):
        try:
            l = ctx()["io"][0]
            if len(l) > self.LINES:
                return
            l.append([kw.get("z", 1), kw.get("x", 1), " ".join(a)])
        except:
            ...

        # self.out("\x1b[{};{}H{}".format(, " ".join(a)))


def step(x=70, y=0, z=2):
    import time

    def box(t, x, y, z):
        lines = t.split("\n")
        fill = "─" * len(t)
        if z > 1:
            print("┌%s┐" % fill, x=x, z=z - 1)
        for t in lines:
            print("│%s│" % t, x=x, z=z)
            z += 1
        print("└%s┘\n" % fill, x=x, z=z)

    with tui() as print:
        # draw a clock
        t = "%02d:%02d:%02d ☢ 99%% " % time.localtime()[3:6]
        box(t, x=x, y=y, z=z)


async def clock(x=70, y=0, z=2):
    # run as a daemon
    while True:  # not asyncio.exit:
        step(x, y, z)
        await asyncio.sleep(1)
        sys.stdout.flush()


def start(x=70, y=0, z=2):
    asyncio.create_task(clock(x, y, z))



plan = []


def register(func, *args, **kwargs):
    global plan
    plan.append(
        (
            func,
            arg,
            kwargs,
        )
    )


def unregister(func):
    global plan
    todel = []
    for i, elem in enumerate(plan):
        if elem[0] is func:
            todel.append(i)

    while len(todel):
        plan.pop(todel.pop())


def exiting():
    global plan
    while len(plan):
        f, argv, kw = plan.pop()
        f(argv, kw)


# replace stock one
import sys

sys.modules["atexit"] = sys.modules["aio.atexit"]



"""

The base class for all messages (including events).
"""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from . import _time
from ._context import active_message_pump
from .case import camel_to_snake

if TYPE_CHECKING:
    from .message_pump import MessagePump
    from .widget import Widget


class Message:
    """Base class for a message."""

    __slots__ = [
        "_sender",
        "time",
        "_forwarded",
        "_no_default_action",
        "_stop_propagation",
        "_prevent",
    ]

    ALLOW_SELECTOR_MATCH: ClassVar[set[str]] = set()
    """Additional attributes that can be used with the [`on` decorator][textual.on].

    These attributes must be widgets.
    """
    bubble: ClassVar[bool] = True  # Message will bubble to parent
    verbose: ClassVar[bool] = False  # Message is verbose
    no_dispatch: ClassVar[bool] = False  # Message may not be handled by client code
    namespace: ClassVar[str] = ""  # Namespace to disambiguate messages
    handler_name: ClassVar[str]
    """Name of the default message handler."""

    def __init__(self) -> None:
        self.__post_init__()

    def __post_init__(self) -> None:
        """Allow dataclasses to initialize the object."""
        self._sender: MessagePump | None = active_message_pump.get(None)
        self.time: float = _time.get_time()
        self._forwarded = False
        self._no_default_action = False
        self._stop_propagation = False
        self._prevent: set[type[Message]] = set()

    def __init_subclass__(
        cls,
        bubble: bool | None = True,
        verbose: bool = False,
        no_dispatch: bool | None = False,
        namespace: str | None = None,
    ) -> None:
        super().__init_subclass__()
        if bubble is not None:
            cls.bubble = bubble
        cls.verbose = verbose
        if no_dispatch is not None:
            cls.no_dispatch = no_dispatch
        if namespace is not None:
            cls.namespace = namespace
        name = camel_to_snake(cls.__name__)
        cls.handler_name = f"on_{namespace}_{name}" if namespace else f"on_{name}"

    @property
    def control(self) -> Widget | None:
        """The widget associated with this message, or None by default."""
        return None

    @property
    def is_forwarded(self) -> bool:
        """Has the message been forwarded?"""
        return self._forwarded

    def _set_forwarded(self) -> None:
        """Mark this event as being forwarded."""
        self._forwarded = True

    def _set_sender(self, sender: MessagePump) -> None:
        """Set the sender."""
        self._sender = sender

    def can_replace(self, message: "Message") -> bool:
        """Check if another message may supersede this one.

        Args:
            message: Another message.

        Returns:
            True if this message may replace the given message
        """
        return False

    def prevent_default(self, prevent: bool = True) -> Message:
        """Suppress the default action(s). This will prevent handlers in any base classes
        from being called.

        Args:
            prevent: True if the default action should be suppressed,
                or False if the default actions should be performed.
        """
        self._no_default_action = prevent
        return self

    def stop(self, stop: bool = True) -> Message:
        """Stop propagation of the message to parent.

        Args:
            stop: The stop flag.
        """
        self._stop_propagation = stop
        return self

    def _bubble_to(self, widget: MessagePump) -> None:
        """Bubble to a widget (typically the parent).

        Args:
            widget: Target of bubble.
        """
        widget.post_message(self)
